#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Union
from .parameters import Parameters
from openqaoa.qaoa_parameters import Hamiltonian


class QaoaParameters(Parameters):
    """
    Parameter class to initialise parameters to run a desired QAOA program.


    Attributes
    ----------
    p: `int`
        Number of QAOA layers.
    backend: `str`
        String containing the specific backend in which QAOA is run.
    params_type: `str`
        String specificying the parametrization of the QAOA ansatz.
    init_type: `str`
        String specifiying the initialization of the QAOA variational parameters.
    shots: `int`
        Number of shots considered for statistics.
    qpu_params: `dict`
        Dictionary containing further specifications of a hardware backend.
    """
    def __init__(self,
                 p: int = 1,
                 backend: str = 'vectorized',
                 params_type: str = 'standard',
                 init_type: str = 'ramp',
                 shots: int = None,
                 optimizer_dict: dict = {'method': 'cobyla', 'maxiter': 200},
                 qpu_params: dict = None):
        """
        Initialises QAOA program parameters.

        Parameters
        ----------
        p: `int`
            Number of QAOA layers. Defaults to 1.
        backend: `str`
            String containing the specific backend in which QAOA is run. Defaults to vectorized.
        params_type: `str`
            String specificying the parametrization of the QAOA ansatz. Defaults to 'standard'.
        init_type: `str`
            String specifiying the initialization of the QAOA variational parameters. Defaults to 'ramp'.
        shots: `int`
            Number of shots considered for statistics. Defaults to None.
        qpu_params: `dict`
            Dictionary containing further specifications of a hardware backend. Defaults to None.
        """
        self.p = p
        self.backend = backend
        self.params_type = params_type
        self.init_type = init_type
        self.shots = shots
        self.optimizer_dict = optimizer_dict
        self.qpu_params = qpu_params


class RqaoaParameters(Parameters):
    """
    Parameter class to initialise parameters to run a desired RQAOA program.

    Attributes
    ----------
    rqaoa_type: `int`
        String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
        'adaptive'.
    n_max: `int`
        Maximum number of eliminations allowed at each step when using the adaptive method.
    steps: `Union[list,int]`
        Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins eliminated
        at each step. If a list is passed, the algorithm will follow the list to select how many spins to eliminate 
        at each step. Note that the list needs enough elements to specify eliminations from the initial number of qubits
        up to the cutoff value. If the list contains more, the algorithm will follow instructions until the cutoff value 
        is reached.
    n_cutoff: `int`
        Cutoff value at which the RQAOA algorithm obtains the solution classically.
    shots: `int`
        Number of shots considered for statistics.
    qpu_params: `dict`
        Dictionary containing further specifications of a hardware backend.
    original_hamiltonian: `Hamiltonian`
        Hamiltonian encoding the original problem fed into the RQAOA algorithm.
    """
    
    def __init__(self,
                 rqaoa_type: str = 'adaptive',
                 n_max: int = 1,
                 steps: Union[list,int] = 1,
                 n_cutoff: int = 5,
                 original_hamiltonian: Hamiltonian = None):
        """
        Initialises RQAOA program parameters.

        Parameters
        ----------
        rqaoa_type: `int`
            String specifying the RQAOA scheme under which eliminations are computed. The two methods are 'custom' and
            'adaptive'. Defaults to 'adaptive'.
        n_max: `int`
            Maximum number of eliminations allowed at each step when using the adaptive method. Defaults to 1.
        steps: `Union[list,int]`
            Elimination schedule for the RQAOA algorithm. If an integer is passed, it sets the number of spins eliminated
            at each step. If a list is passed, the algorithm will follow the list to select how many spins to eliminate 
            at each step. Note that the list needs enough elements to specify eliminations from the initial number of qubits
            up to the cutoff value. If the list contains more, the algorithm will follow instructions until the cutoff value 
            is reached. Defaults to 1.
        n_cutoff: `int`
            Cutoff value at which the RQAOA algorithm obtains the solution classically. Defaults to 5.
        original_hamiltonian: `Hamiltonian`
            Hamiltonian encoding the original problem fed into the RQAOA algorithm. Defaults to None.
        """
        self.rqaoa_type = rqaoa_type
        self.n_max = n_max
        self.steps = steps
        self.n_cutoff = n_cutoff
        self.original_hamiltonian = original_hamiltonian



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Conversion functions for the different QAOA Parametrizations. So far we only
only support going from less to more specialiced parametrizations. The type
tree looks as follows:
   Extended   <--------- FourierExtended
       ^                      ^
       |                      |
StandardWithBias <------ FourierWithBias
       ^                      ^
       |                      |
    Standard  <----------- Fourier
       ^
       |
    Annealing
"""
from __future__ import annotations
from typing import Type

from copy import deepcopy
import numpy as np
from scipy.fftpack import dct, dst

from . import (annealingparams,
               standardparams,
               fourierparams,
               extendedparams)

QAOAVariationalAnnealingParams = annealingparams.QAOAVariationalAnnealingParams
QAOAVariationalStandardParams = standardparams.QAOAVariationalStandardParams
QAOAVariationalExtendedParams = extendedparams.QAOAVariationalExtendedParams
QAOAVariationalStandardWithBiasParams = standardparams.QAOAVariationalStandardWithBiasParams
QAOAVariationalFourierParams = fourierparams.QAOAVariationalFourierParams
QAOAVariationalFourierWithBiasParams = fourierparams.QAOAVariationalFourierWithBiasParams
QAOAVariationalFourierExtendedParams = fourierparams.QAOAVariationalFourierExtendedParams


def annealing_to_standard(
		params: QAOAVariationalAnnealingParams) -> QAOAVariationalStandardParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardParams
    out.betas = params._annealing_time * (1 - params.schedule) / params.p
    out.gammas = params._annealing_time * params.schedule / params.p

    # and clean up after us
    del out.__schedule
    del out._annealing_time

    return out


def standard_to_standard_w_bias(
        params: QAOAVariationalStandardParams) -> QAOAVariationalStandardWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardWithBiasParams
    out.gammas_singles = params.gammas
    out.gammas_pairs = params.gammas

    # and clean up after us
    del out.__gammas

    return out


def standard_w_bias_to_extended(
		params: QAOAVariationalStandardWithBiasParams) -> QAOAVariationalExtendedParams:
	out = deepcopy(params)
	out.__class__ = QAOAVariationalExtendedParams
	out.betas_singles = np.outer(params.betas, np.ones(len(params.mixer_1q_coeffs)))
	out.betas_pairs = np.outer(params.betas, np.ones(len(params.mixer_2q_coeffs)))

	out.gammas_singles = np.outer(params.gammas_singles,
									np.ones(len(params.cost_1q_coeffs)))
	out.gammas_pairs = np.outer(params.gammas_pairs,
								np.ones(len(params.cost_2q_coeffs)))
	return out


def fourier_to_standard(
	params: QAOAVariationalFourierParams) -> QAOAVariationalStandardParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardParams
    out.betas = dct(params.v, n=out.p)
    out.gammas = dst(params.u, n=out.p)

    # and clean up
    del out.__u
    del out.__v
    del out.q

    return out


def fourier_w_bias_to_standard_w_bias(
        params: QAOAVariationalFourierWithBiasParams) -> QAOAVariationalStandardWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalStandardWithBiasParams
    out.betas = dct(params.v, n=out.p)
    out.gammas_singles = dst(params.u_singles, n=out.p)
    out.gammas_pairs = dst(params.u_pairs, n=out.p)

    # and clean up
    del out.__u_singles
    del out.__u_pairs
    del out.__v
    del out.q

    return out


def fourier_to_fourier_w_bias(
		params: QAOAVariationalFourierParams) -> QAOAVariationalFourierWithBiasParams:
    out = deepcopy(params)
    out.__class__ = QAOAVariationalFourierWithBiasParams
    out.u_singles = params.u
    out.u_pairs = params.u

    # and clean up
    del out.__u

    return out


def fourier_w_bias_to_fourier_extended(
		params: QAOAVariationalFourierWithBiasParams) -> QAOAVariationalFourierExtendedParams:
	out = deepcopy(params)
	out.__class__ = QAOAVariationalFourierExtendedParams
	out.v_singles = np.outer(params.v, np.ones(len(params.mixer_1q_coeffs)))
	out.v_pairs = np.outer(params.v, np.ones(len(params.mixer_2q_coeffs)))
	out.u_singles = np.outer(params.u_singles,
								np.ones(len(params.cost_1q_coeffs)))
	out.u_pairs = np.outer(params.u_pairs,
							np.ones(len(params.cost_2q_coeffs)))

	del out.__v

	return out


def fourier_extended_to_extended(
        params: QAOAVariationalFourierExtendedParams) -> QAOAVariationalExtendedParams:
	out = deepcopy(params)
	out.__class__ = QAOAVariationalExtendedParams
	out.betas_singles = dct(params.v_singles, n=out.p, axis=0)
	out.betas_pairs = dct(params.v_pairs, n=out.p, axis=0)
	out.gammas_singles = dst(params.u_singles, n=out.p, axis=0)
	out.gammas_pairs = dst(params.u_pairs, n=out.p, axis=0)

    # and clean up
	del out.__u_singles
	del out.__u_pairs
	del out.__v_singles
	del out.__v_pairs
	del out.q

	return out


# #############################################################################
# And now all the possible compositions as well:
# Todo: Create this code automatically by traversing the tree?
# #############################################################################

def annealing_to_standard_w_bias(
        params: QAOAVariationalAnnealingParams) -> QAOAVariationalStandardWithBiasParams:
    return standard_to_standard_w_bias(annealing_to_standard(params))


def annealing_to_extended(
        params: QAOAVariationalAnnealingParams) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(annealing_to_standard_w_bias(params))


def standard_to_extended(
        params: QAOAVariationalStandardParams) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(standard_to_standard_w_bias(params))


def fourier_to_fourier_extended(
        params: QAOAVariationalFourierParams) -> QAOAVariationalFourierExtendedParams:
    return fourier_w_bias_to_fourier_extended(
                fourier_to_fourier_w_bias(params))


def fourier_to_standard_w_bias(
        params: QAOAVariationalFourierParams) -> QAOAVariationalStandardWithBiasParams:
    return standard_to_standard_w_bias(fourier_to_standard(params))


def fourier_to_extended(
        params: QAOAVariationalFourierParams) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(fourier_to_standard_w_bias(params))


def fourier_w_bias_to_extended(
        params: QAOAVariationalFourierWithBiasParams) -> QAOAVariationalExtendedParams:
    return standard_w_bias_to_extended(
            fourier_w_bias_to_standard_w_bias(params))


# A dict with all the conversion functions accessible by
# (input_type, output_type)
conversion_functions =\
    {
     (QAOAVariationalFourierExtendedParams, QAOAVariationalExtendedParams): fourier_extended_to_extended,
     (QAOAVariationalFourierWithBiasParams, QAOAVariationalStandardWithBiasParams):\
     fourier_w_bias_to_standard_w_bias,
     (QAOAVariationalFourierParams, QAOAVariationalStandardParams): fourier_to_standard,
     (QAOAVariationalAnnealingParams, QAOAVariationalStandardParams): annealing_to_standard,
     (QAOAVariationalStandardParams, QAOAVariationalStandardWithBiasParams): standard_to_standard_w_bias,
     (QAOAVariationalStandardWithBiasParams, QAOAVariationalExtendedParams): standard_w_bias_to_extended,
     (QAOAVariationalFourierParams, QAOAVariationalFourierWithBiasParams): fourier_to_fourier_w_bias,
     (QAOAVariationalFourierWithBiasParams, QAOAVariationalFourierExtendedParams):\
     fourier_w_bias_to_fourier_extended,
     (QAOAVariationalAnnealingParams, QAOAVariationalStandardWithBiasParams): annealing_to_standard_w_bias,
     (QAOAVariationalAnnealingParams, QAOAVariationalExtendedParams): annealing_to_extended,
     (QAOAVariationalStandardParams, QAOAVariationalExtendedParams): standard_to_extended,
     (QAOAVariationalFourierParams, QAOAVariationalFourierExtendedParams): fourier_to_fourier_extended,
     (QAOAVariationalFourierParams, QAOAVariationalStandardWithBiasParams): fourier_to_standard_w_bias,
     (QAOAVariationalFourierParams, QAOAVariationalExtendedParams): fourier_to_extended,
     (QAOAVariationalFourierWithBiasParams, QAOAVariationalExtendedParams): fourier_w_bias_to_extended,
    }


def converter(params, target_type: type):
    """
    Convert ``params`` to type ``target_type``
    Parameters
    ----------
    params:
        The input parameters
    target_type:
        The target type
    Returns
    -------
    target_type:
        The converted parameters
    Raises
    ------
    TypeError:
        If conversion from type(params) to target_type is not supported.
    """
    try:
        out = conversion_functions[(type(params), target_type)](params)
        return out
    except KeyError:
        raise TypeError(f"Conversion from {type(params)} to {target_type} "
                        "not supported.")


#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from logging.config import dictConfig
from re import I
import matplotlib.pyplot as plt
from typing import Type
import numpy as np

from .logger_vqa import Logger
from ..qaoa_parameters.operators import Hamiltonian
from ..utilities import qaoa_probabilities, bitstring_energy


def most_probable_bitstring(cost_hamiltonian, measurement_outcomes):

    mea_out = list(measurement_outcomes.values())
    index_likliest_states = np.argwhere(mea_out == np.max(mea_out))
    # degeneracy = len(index_likliest_states)
    solutions_bitstrings = [list(measurement_outcomes.keys())[
        e[0]] for e in index_likliest_states]

    return {'solutions_bitstrings': solutions_bitstrings,
            'bitstring_energy': bitstring_energy(cost_hamiltonian, solutions_bitstrings[0])}


class Result():
    '''
    A class to handle the results of QAOA workflows

    Parameters
    ----------
    log: `Logger`
        The raw logger generated from the training vqa part of the QAOA. 
    method: `str`
        Stores the name of the optimisation used by the classical optimiser
    '''

    def __init__(self,
                 log: Type[Logger],
                 method: Type[str],
                 cost_hamiltonian: Type[Hamiltonian]):

        self.method = method

        self.evals = {
            'number of evals': log.func_evals.best[0],
            'jac evals': log.jac_func_evals.best[0],
            'qfim evals': log.qfim_func_evals.best[0]
        }

        self.intermediate = {
            'angles log': np.array(log.param_log.history).tolist(),
            'intermediate cost': log.cost.history,
            'intermediate measurement outcomes':
                log.measurement_outcomes.history
        }

        self.optimized = {
            'optimized angles': np.array(log.param_log.best[0]).tolist(),
            'optimized cost': log.cost.best[0],
            'optimized measurement outcomes':
                log.measurement_outcomes.best[0]
                if log.measurement_outcomes.best != [] else {}
        }

        self.most_probable_states = most_probable_bitstring(cost_hamiltonian,
                                                            self.get_counts(log.measurement_outcomes.best[0]))


    # def __repr__(self):
    #     """Return an overview over the parameters and hyperparameters
    #     Todo
    #     ----
    #     Split this into ``__repr__`` and ``__str__`` with a more verbose
    #     output in ``__repr__``.
    #     """
    #     string = "Optimization Results:\n"
    #     string += "\tThe solution is " + str(self.solution['degeneracy']) + " degenerate" "\n"
    #     string += "\tThe most probable bitstrings are: " + str(self.solution['bitstring']) + "\n"
    #     string += "\tThe associated cost is: " + str(self.optimized['optimized cost']) + "\n"

    #     return (string)

    @staticmethod
    def get_counts(measurement_outcomes):
        """
        Converts probabilities to counts when the measurement outcomes are a numpy array, that is a state vector

        Parameters
        ----------
        measurement_outcomes: `Union[np.array, dict]`
            The measurement outcome as returned by the Logger. It can either be a statevector or a count dictionary
        Returns
        -------
        `dict`
            The count dictionary obtained either throught the statevector or the actual measurement counts
        """

        if type(measurement_outcomes) == type(np.array([])):
            measurement_outcomes = qaoa_probabilities(measurement_outcomes)

        return measurement_outcomes

    def plot_cost(self, figsize=(10, 8), label='Cost', linestyle='--', color='b', ax=None):
        """
        A simpler helper function to plot the cost associated to a QAOA workflow

        Parameters
        ----------
        figsize: `tuple`
            The size of the figure to be plotted. Defaults to (10,8).
        label: `str`
            The label of the cost line, defaults to 'Cost'.
        linestyle: `str`
            The linestyle of the poloit. Defaults to '--'.
        color: `str`
            The color of the line. Defaults to 'b'.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        
        ax.plot(range(self.evals['number of evals'] - self.evals['jac evals'] - self.evals['qfim evals']),
                    self.intermediate['intermediate cost'],
                    label=label,
                    linestyle=linestyle,
                    color=color)

        ax.set_ylabel('Cost')
        ax.set_xlabel('Number of function evaluations')
        ax.legend()
        ax.set_title('Cost history')

        return



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
from typing import List, Union, Tuple, Any, Callable, Iterable, Type
import numpy as np

from .operators import Hamiltonian
from .hamiltonianmapper import HamiltonianMapper



def _is_iterable_empty(in_iterable):
    if isinstance(in_iterable, Iterable):    # Is Iterable
        return all(map(_is_iterable_empty, in_iterable))
    return False    # Not an Iterable


class shapedArray:
    """Decorator-Descriptor for arrays that have a fixed shape.

    This is used to facilitate automatic input checking for all the different
    internal parameters. Each instance of this class can be removed without
    replacement and the code should still work, provided the user provides
    only correct angles to below parameter classes

    Parameters
    ----------
    shape: Callable[[Any], Tuple]:
        Returns the shape for self.values

    Example
    -------
    With this descriptor, the two following are equivalent:

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)
                self._my_attribute = None

            @property
            def my_attribute(self):
                return _my_attribute

            @my_attribute.setter
            def my_attribute(self):
                try:
                    self._my_attribute = np.reshape(values, self.shape)
                except ValueError:
                    raise ValueError("my_attribute must have shape "
                                    f"{self.shape}")


    can be simplified to

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)

            @shapedArray
            def my_attribute(self):
                return self.shape
    """

    def __init__(self, shape: Callable[[Any], Tuple]):
        """The constructor. See class documentation"""
        self.name = shape.__name__
        self.shape = shape

    def __set__(self, obj, values):
        """The setter with input checking."""
        try:
            setattr(obj, f"__{self.name}", np.reshape(values, self.shape(obj)))
        except ValueError:
            raise ValueError(f"{self.name} must have shape {self.shape(obj)}")

    def __get__(self, obj, objtype):
        """The getter."""
        return getattr(obj, f"__{self.name}")


class VQACircuitParams(ABC):
    """
    Parameters class to construct a specific quantum ansatz to attack
    a problem

    Parameters
    ----------
    algorithm: `str`

    Attributes
    ----------
    algorithm: `str`
    """

    def __init__(self,
                 algorithm: str):
        self.algorithm = algorithm


class QAOACircuitParams(VQACircuitParams):
    """
    Create the problem attributes consisting of the Hamiltonian and Hamiltonian
    squared as dictionaries, QAOA 'p' value and other specific parameters.

    Parameters
    ----------    
    cost_hamiltonian: `Hamiltonian` 

    mixer_hamiltonian: `Hamiltonian`

    p: 
        Number of QAOA layers; defaults to 1 if not specified

    Attributes
    ----------
        All the parameters specified above are initialised as attributes in __init__

    Properties
    ----------
    hamiltonian: 
        Returns the problem Hamiltonian as a dictionary
    hamiltonian_squared: 
        Returns the problem Hamiltonian Squared as a dictionary
    """

    def __init__(self,
                 cost_hamiltonian: Hamiltonian,
                 mixer_hamiltonian: Hamiltonian,
                 p: int):

        super().__init__(algorithm='QAOA')
        self.cost_hamiltonian = cost_hamiltonian
        self.mixer_hamiltonian = mixer_hamiltonian
        self.qureg = cost_hamiltonian.qureg
        self.p = p

    def __repr__(self):
        """Return an overview over the parameters and hyperparameters

        Todo
        ----
        Split this into ``__repr__`` and ``__str__`` with a more verbose
        output in ``__repr__``.
        """
        string = "Circuit Parameters:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tregister: " + str(self.qureg) + "\n" + "\n"

        string += "Cost Hamiltonian:\n"
        string += "\tcost_qubits_singles: " + \
            str(self.cost_hamiltonian.qubits_singles) + "\n"
        string += "\tcost_single_qubit_coeffs: " + \
            str(self.cost_hamiltonian.single_qubit_coeffs) + "\n"
        string += "\tcost_qubits_pairs: " + \
            str(self.cost_hamiltonian.qubits_pairs) + "\n"
        string += "\tcost_pair_qubit_coeffs: " + \
            str(self.cost_hamiltonian.pair_qubit_coeffs) + "\n" + "\n"

        string += "Mixer Hamiltonian:\n"
        string += "\tmixer_qubits_singles: " + \
            str(self.mixer_hamiltonian.qubits_singles) + "\n"
        string += "\tmixer_single_qubit_coeffs: " + \
            str(self.mixer_hamiltonian.single_qubit_coeffs) + "\n"
        string += "\tmixer_qubits_pairs: " + \
            str(self.mixer_hamiltonian.qubits_pairs) + "\n"
        string += "\tmixer_pair_qubit_coeffs: " + \
            str(self.mixer_hamiltonian.pair_qubit_coeffs) + "\n"

        return string

    @property
    def abstract_circuit(self) -> List:

        cost_gate_map_list = HamiltonianMapper.repeat_gate_maps(
            self.cost_hamiltonian, 'cost', self.p)
        mixer_gate_map_list = HamiltonianMapper.repeat_gate_maps(
            self.mixer_hamiltonian, 'mixer', self.p)

        _abstract_circuit = []
        for each_p in range(self.p):
            _abstract_circuit.extend(cost_gate_map_list[each_p])
            _abstract_circuit.extend(mixer_gate_map_list[each_p])

        return _abstract_circuit


class QAOAVariationalBaseParams(ABC):
    """
    A class that initialises and keeps track of the Variational
    parameters 

    Parameters
    ----------
    qaoa_circuit_params:
        Specify the circuit parameters to construct circuit angles to be 
        used for training

    Attributes
    ----------

    """

    def __init__(self, qaoa_circuit_params: Type[QAOACircuitParams]):

        self.qaoa_circuit_params = qaoa_circuit_params
        self.p = self.qaoa_circuit_params.p
        self.cost_1q_coeffs = qaoa_circuit_params.cost_hamiltonian.single_qubit_coeffs
        self.cost_2q_coeffs = qaoa_circuit_params.cost_hamiltonian.pair_qubit_coeffs
        self.mixer_1q_coeffs = qaoa_circuit_params.mixer_hamiltonian.single_qubit_coeffs
        self.mixer_2q_coeffs = qaoa_circuit_params.mixer_hamiltonian.pair_qubit_coeffs

    def __len__(self):
        """
        Returns
        -------
        int:
                the length of the data produced by self.raw() and accepted by
                self.update_from_raw()
        """
        raise NotImplementedError()

    def __repr__(self):

        raise NotImplementedError()

    def __str__(self):

        return self.__repr__()

    @property
    def mixer_1q_angles(self) -> np.ndarray:
        """2D array with the X-rotation angles.

        1st index goes over p and the 2nd index over the qubits to
        apply X-rotations on. These are needed by ``qaoa.cost_function.make_qaoa_memory_map``
        """
        raise NotImplementedError()

    @property
    def mixer_2q_angles(self) -> np.ndarray:
        """2D array with the X-rotation angles.

        1st index goes over p and the 2nd index over the qubits to
        apply X-rotations on. These are needed by ``qaoa.cost_function.make_qaoa_memory_map``
        """
        raise NotImplementedError()

    @property
    def cost_1q_angles(self) -> np.ndarray:
        """2D array with the ZZ-rotation angles.

        1st index goes over the p and the 2nd index over the qubit
        pairs, to apply ZZ-rotations on. These are needed by ``qaoa.cost_function.make_qaoa_memory_map``
        """
        raise NotImplementedError()

    @property
    def cost_2q_angles(self) -> np.ndarray:
        """2D array with Z-rotation angles.

        1st index goes over the p and the 2nd index over the qubit
        pairs, to apply Z-rotations on. These are needed by ``qaoa.cost_function.make_qaoa_memory_map``
        """
        raise NotImplementedError()

    def update_from_raw(self, new_values: Union[list, np.array]):
        """
        Update all the parameters from a 1D array.

        The input has the same format as the output of ``self.raw()``.
        This is useful for ``scipy.optimize.minimize`` which expects
        the parameters that need to be optimized to be a 1D array.

        Parameters
        ----------
        new_values:
                A 1D array with the new parameters. Must have length  ``len(self)``
                and the ordering of the flattend ``parameters`` in ``__init__()``.

        """
        raise NotImplementedError()

    def raw(self) -> np.ndarray:
        """
        Return the parameters in a 1D array.

        This 1D array is needed by ``scipy.optimize.minimize`` which expects
        the parameters that need to be optimized to be a 1D array.

        Returns
        -------
        np.array:
                The parameters in a 1D array. Has the same output
                format as the expected input of ``self.update_from_raw``. Hence
                corresponds to the flattened `parameters` in `__init__()`

        """
        raise NotImplementedError()

    @classmethod
    def linear_ramp_from_hamiltonian(cls,
                                     qaoa_circuit_params: QAOACircuitParams,
                                     time: float = None):
        """Alternative to ``__init__`` that already fills ``parameters``.

        Calculate initial parameters from register, terms, weights (specifiying a Hamiltonian), corresponding to a
        linear ramp annealing schedule and return a ``QAOAParams`` object.

        Parameters
        ----------
        qaoa_circuit_params:
                QAOACircuitParams object containing information about terms,weights,register and p

        time:
                Total annealing time. Defaults to ``0.7*p``.

        Returns
        -------
        VariationalBaseParams
                The initial parameters for a linear ramp for ``hamiltonian``.

        """
        raise NotImplementedError()

    @classmethod
    def random(cls, qaoa_circuit_params: QAOACircuitParams, seed: int = None):
        """
        Initialise parameters randomly

        Parameters
        ----------
        qaoa_circuit_params:
                QAOACircuitParams object containing information about terms,weights,register and p

        seed:
                Use a fixed seed for reproducible random numbers

        Returns
        -------
        VariationalBaseParams
                Randomly initialiased parameters
        """
        raise NotImplementedError()

    @classmethod
    def empty(cls, qaoa_circuit_params: QAOACircuitParams):
        """
        Alternative to ``__init__`` that only takes ``qaoa_circuit_params`` and
        fills ``parameters`` via ``np.empty``

        Parameters
        ----------
        qaoa_circuit_params:
            QAOACircuitParams object containing information about terms,weights,register and p

        Returns
        -------
        VariationalBaseParams:
            A Parameter object with the parameters filled by ``np.empty``
        """
        raise NotImplementedError()

    @classmethod
    def from_other_parameters(cls, params):
        """Alternative to ``__init__`` that takes parameters with less degrees
        of freedom as the input.
        Parameters
        ----------
        params: QAOAVaritionalBaseParams
            The input parameters object to construct the new parameters object from.
        Returns
        -------
        QAOAVariationalBaseParams:
            The converted paramters s.t. all the rotation angles of the in
            and output parameters are the same.
        """
        from . import converter
        return converter(params, cls)

    def raw_rotation_angles(self) -> np.ndarray:
        """
        Flat array of the rotation angles for the memory map for the
        parametric circuit.

        Returns
        -------
        np.array:
                Returns all single rotation angles in the ordering
                ``(x_rotation_angles, gamma_singles, zz_rotation_angles)`` where
                ``x_rotation_angles = (beta_q0_t0, beta_q1_t0, ... , beta_qn_tp)``
                and the same for ``z_rotation_angles`` and ``zz_rotation_angles``

        """
        raw_data = np.concatenate((self.mixer_1q_angles.flatten(),
                                   self.mixer_2q_angles.flatten(),
                                   self.cost_1q_angles.flatten(),
                                   self.cost_1q_angles.flatten(),))
        return raw_data

    def plot(self, ax=None, **kwargs):
        """
        Plots ``self`` in a sensible way to the canvas ``ax``, if provided.

        Parameters
        ----------
        ax: matplotlib.axes._subplots.AxesSubplot
                The canvas to plot itself on
        kwargs:
                All remaining keyword arguments are passed forward to the plot
                function

        """
        raise NotImplementedError()

    # def QAOA_circuit_params(self):
    # 	"""
    # 	Returns parameters needed for QAOA circuit in tuple form.
    # 	"""

    # 	p = self.p
    # 	x_rotation_angles = self.x_rotation_angles
    # 	bias_qubits = self.qubits_singles
    # 	pairs = self.qubits_pairs
    # 	z_rotation_angles = self.z_rotation_angles
    # 	zz_rotation_angles = self.zz_rotation_angles

    # 	return (p,x_rotation_angles,bias_qubits,z_rotation_angles,pairs,zz_rotation_angles)
    # 	return (self.p, self.mixer_1q_angles, self.mixer_2q_angles,
    # 			self.cost_1q_angles, self.cost_2q_angles)


class QAOAParameterIterator:
    """An iterator to sweep one parameter over a range in a QAOAParameter object.

    Parameters
    ----------
    qaoa_params:
        The initial QAOA parameters, where one of them is swept over
    the_parameter:
        A string specifying, which parameter should be varied. It has to be
        of the form ``<attr_name>[i]`` where ``<attr_name>`` is the name
        of the _internal_ list and ``i`` the index, at which it sits. E.g.
        if ``qaoa_params`` is of type ``AnnealingParams``
        and  we want to vary over the second timestep, it is
        ``the_parameter = "times[1]"``.
    the_range:
        The range, that ``the_parameter`` should be varied over

    Todo
    ----
    - Add checks, that the number of indices in ``the_parameter`` matches
      the dimensions of ``the_parameter``
    - Add checks, that the index is not too large

    Example
    -------
    Assume qaoa_params is of type ``StandardWithBiasParams`` and
    has `p >= 2`. Then the following code produces a loop that
    sweeps ``gammas_singles[1]`` over the range ``(0, 1)`` in 4 layers:

    .. code-block:: python

        the_range = np.arange(0, 1, 0.4)
        the_parameter = "gammas_singles[1]"
        param_iterator = QAOAParameterIterator(qaoa_params, the_parameter, the_range)
        for params in param_iterator:
            # do what ever needs to be done.
            # we have type(params) == type(qaoa_params)
    """

    def __init__(self,
                 variational_params: QAOAVariationalBaseParams,
                 the_parameter: str,
                 the_range: Iterable[float]):
        """See class documentation for details"""
        self.params = variational_params
        self.iterator = iter(the_range)
        self.the_parameter, *indices = the_parameter.split("[")
        indices = [i.replace(']', '') for i in indices]
        if len(indices) == 1:
            self.index0 = int(indices[0])
            self.index1 = False
        elif len(indices) == 2:
            self.index0 = int(indices[0])
            self.index1 = int(indices[1])
        else:
            raise ValueError("the_parameter has to many indices")

    def __iter__(self):
        return self

    def __next__(self):
        # get next value from the_range
        value = next(self.iterator)

        # 2d list or 1d list?
        if self.index1 is not False:
            getattr(self.params, self.the_parameter)[
                self.index0][self.index1] = value
        else:
            getattr(self.params, self.the_parameter)[self.index0] = value

        return self.params



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Union
import matplotlib.pyplot as plt
import numpy as np
from .baseparams import QAOACircuitParams, QAOAVariationalBaseParams, shapedArray, _is_iterable_empty


class QAOAVariationalStandardParams(QAOAVariationalBaseParams):
    r"""
    QAOA parameters that implement a state preparation circuit with

    .. math::

        e^{-i \beta_p H_0}
        e^{-i \gamma_p H_c}
        \cdots
        e^{-i \beta_0 H_0}
        e^{-i \gamma_0 H_c}

    This corresponds to the parametrization used by Farhi in his
    original paper [https://arxiv.org/abs/1411.4028]

    Parameters
    ----------
    hyperparameters:
        The hyperparameters containing the register, terms, weights, and the number of layers
        ``hyperparameters = (register, terms, weights, p)``
    parameters:
        Tuple containing ``(betas, gammas)`` with dimensions
        ``(p, p)``

    Attributes
    ----------
    betas: np.array
        1D array with the betas from above
    gammas: np.array
        1D array with the gamma from above
    """

    def __init__(self,
                 qaoa_circuit_params: QAOACircuitParams,
                 betas: List[Union[float, int]],
                 gammas: List[Union[float, int]]):
        # setup reg, qubits_singles and qubits_pairs
        super().__init__(qaoa_circuit_params)
        self.betas = np.array(betas)
        self.gammas = np.array(gammas)

    def __repr__(self):
        string = "Standard Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Variational Parameters:\n"
        string += "\tbetas: " + str(self.betas) + "\n"
        string += "\tgammas: " + str(self.gammas) + "\n"
        return(string)

    def __len__(self):
        return self.p * 2

    @shapedArray
    def betas(self):
        return self.p

    @shapedArray
    def gammas(self):
        return self.p

    @property
    def mixer_1q_angles(self):
        return 2*np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2*np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2*np.outer(self.gammas, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2*np.outer(self.gammas, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite self.betas with new ones
        self.betas = np.array(new_values[0:self.p])
        new_values = new_values[self.p:]    # cut betas from new_values
        self.gammas = np.array(new_values[0:self.p])
        new_values = new_values[self.p:]

        if len(new_values) != 0:
            raise RuntimeWarning("Incorrect dimension specified for new_values"
                                 "to construct the new betas and new gammas")

    def raw(self):
        raw_data = np.concatenate((self.betas, self.gammas))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(cls,
                                     qaoa_circuit_params: QAOACircuitParams,
                                     time: float = None):
        """
        Returns
        -------
        StandardParams
            A ``StandardParams`` object with parameters according
            to a linear ramp schedule for the Hamiltonian specified by register, terms, weights.
        """
        p = qaoa_circuit_params.p

        if time is None:
            time = float(0.7 * p)
        # create evenly spaced timelayers at the centers of p intervals
        dt = time / p
        # fill betas, gammas_singles and gammas_pairs
        betas = np.linspace((dt / time) * (time * (1 - 0.5 / p)),
                            (dt / time) * (time * 0.5 / p), p)
        gammas = betas[::-1]
        # wrap it all nicely in a qaoa_parameters object
        params = cls(qaoa_circuit_params, betas, gammas)

        return params

    @classmethod
    def random(cls, qaoa_circuit_params: QAOACircuitParams, seed: int = None):
        """
        Returns
        -------
        StandardParams
            Randomly initialised ``StandardParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        betas = np.random.uniform(0, np.pi, qaoa_circuit_params.p)
        gammas = np.random.uniform(0, np.pi, qaoa_circuit_params.p)

        params = cls(qaoa_circuit_params, betas, gammas)

        return params

    @classmethod
    def empty(cls, qaoa_circuit_params: QAOACircuitParams):
        """
        Initialise Standard Variational params with empty arrays
        """
        p = qaoa_circuit_params.p
        betas = np.empty(p)
        gammas = np.empty(p)

        return cls(qaoa_circuit_params, betas, gammas)

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(self.betas, label="betas", marker="s", ls="", **kwargs)
        ax.plot(self.gammas, label="gammas", marker="^", ls="", **kwargs)
        ax.set_xlabel("timestep", fontsize=12)
        ax.legend()
        
    def convert_to_ext(self, args_std):
        """
        Method that converts a list of parameters in the standard parametrisation form (args_std) to an equivalent list of parameters in the extended parametrisation form.
        
        PARAMETERS
        ----------
        args_std : 
            Parameters (a list of float) in the standard parametrisation form. 

        RETURNS
        -------
        args_ext:
            Parameters (a list of float) in the extended parametrisation form. 
        
        """

        terms_lst = [len(self.mixer_1q_coeffs), len(self.mixer_2q_coeffs), len(self.cost_1q_coeffs), len(self.cost_2q_coeffs)]
        terms_lst_p = np.repeat(terms_lst, [self.p]*len(terms_lst))
        args_ext = []
        for i in range(4):  # 4 types of terms
            for j in range(self.p):
                for k in range(terms_lst_p[i*self.p + j]):
                    if i < 2:
                        args_ext.append(args_std[j])
                    else:
                        args_ext.append(
                            args_std[j + int(len(args_std)/2)])

        return args_ext


class QAOAVariationalStandardWithBiasParams(QAOAVariationalBaseParams):
    r"""
    QAOA parameters that implement a state preparation circuit with

    .. math::

        e^{-i \beta_p H_0}
        e^{-i \gamma_{\textrm{singles}, p} H_{c, \textrm{singles}}}
        e^{-i \gamma_{\textrm{pairs}, p} H_{c, \textrm{pairs}}}
        \cdots
        e^{-i \beta_0 H_0}
        e^{-i \gamma_{\textrm{singles}, 0} H_{c, \textrm{singles}}}
        e^{-i \gamma_{\textrm{pairs}, 0} H_{c, \textrm{pairs}}}

    where the cost hamiltonian is split into :math:`H_{c, \textrm{singles}}`
    the bias terms, that act on only one qubit, and
    :math:`H_{c, \textrm{pairs}}` the coupling terms, that act on two qubits.

    Parameters
    ----------
    hyperparameters:
        The hyperparameters containing the register, terms, weights, and the number of layers
        ``hyperparameters = (register, terms, weights, p)``
    parameters:
        Tuple containing ``(betas, gammas_singles, gammas_pairs)`` with
        dimensions ``(p, p, p)``

    Attributes
    ----------
    betas: np.array
        A 1D array containing the betas from above for each timestep
    gammas_pairs: np.array
        A 1D array containing the gammas_singles from above for each timestep
    gammas_singles: np.array
        A 1D array containing the gammas_pairs from above for each timestep
    """

    def __init__(self,
                 qaoa_circuit_params: QAOACircuitParams,
                 betas: List[Union[float, int]],
                 gammas_singles: List[Union[float, int]],
                 gammas_pairs: List[Union[float, int]]):

        super().__init__(qaoa_circuit_params)
        if not self.cost_1q_coeffs or not self.cost_2q_coeffs:
            raise RuntimeError(f"Please choose {type(self).__name__} parameterisation for "
                               "problems containing both Cost One-Qubit and Two-Qubit terms")

        self.betas = np.array(betas)
        self.gammas_singles = np.array(gammas_singles)
        self.gammas_pairs = np.array(gammas_pairs)

    def __repr__(self):
        string = "Standard with Bias Parameterisation:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "Variational Parameters:\n"
        string += "\tbetas: " + str(self.betas) + "\n"
        string += "\tgammas_singles: " + str(self.gammas_singles) + "\n"
        string += "\tgammas_pairs: " + str(self.gammas_pairs) + "\n"
        return(string)

    def __len__(self):
        return self.p * 3

    @shapedArray
    def betas(self):
        return self.p

    @shapedArray
    def gammas_singles(self):
        return self.p

    @shapedArray
    def gammas_pairs(self):
        return self.p

    @property
    def mixer_1q_angles(self):
        return 2*np.outer(self.betas, self.mixer_1q_coeffs)

    @property
    def mixer_2q_angles(self):
        return 2*np.outer(self.betas, self.mixer_2q_coeffs)

    @property
    def cost_1q_angles(self):
        return 2*np.outer(self.gammas_singles, self.cost_1q_coeffs)

    @property
    def cost_2q_angles(self):
        return 2*np.outer(self.gammas_pairs, self.cost_2q_coeffs)

    def update_from_raw(self, new_values):
        # overwrite self.betas with new ones
        self.betas = np.array(new_values[0:self.p])
        new_values = new_values[self.p:]    # cut betas from new_values
        self.gammas_singles = np.array(new_values[0:self.p])
        new_values = new_values[self.p:]
        self.gammas_pairs = np.array(new_values[0:self.p])
        new_values = new_values[self.p:]

        if len(new_values) != 0:
            raise RuntimeWarning("Incorrect dimension specified for new_values"
                                 "to construct the new betas and new gammas")

    def raw(self):
        raw_data = np.concatenate((self.betas,
                                   self.gammas_singles,
                                   self.gammas_pairs))
        return raw_data

    @classmethod
    def linear_ramp_from_hamiltonian(cls,
                                     qaoa_circuit_params: QAOACircuitParams,
                                     time: float = None):
        """
        Returns
        -------
        StandardParams
            A ``StandardParams`` object with parameters according
            to a linear ramp schedule for the Hamiltonian specified by register, terms, weights.
        """
        p = qaoa_circuit_params.p

        if time is None:
            time = float(0.7 * p)
        # create evenly spaced timelayers at the centers of p intervals
        dt = time / p
        # fill betas, gammas_singles and gammas_pairs
        betas = np.linspace((dt / time) * (time * (1 - 0.5 / p)),
                            (dt / time) * (time * 0.5 / p), p)
        gammas_singles = betas[::-1]
        gammas_pairs = betas[::-1]

        params = cls(qaoa_circuit_params, betas, gammas_singles, gammas_pairs)

        return params

    @classmethod
    def random(cls, qaoa_circuit_params: QAOACircuitParams, seed: int = None):
        """
        Returns
        -------
        StandardParams
            Randomly initialised ``StandardParams`` object
        """
        if seed is not None:
            np.random.seed(seed)

        betas = np.random.uniform(0, np.pi, qaoa_circuit_params.p)
        gammas_singles = np.random.uniform(0, np.pi, qaoa_circuit_params.p)
        gammas_pairs = np.random.uniform(0, np.pi, qaoa_circuit_params.p)

        params = cls(qaoa_circuit_params, betas, gammas_singles, gammas_pairs)
        return params

    @classmethod
    def empty(cls, qaoa_circuit_params: QAOACircuitParams):
        """
        Initialise Standard Variational params with empty arrays
        """
        p = qaoa_circuit_params.p
        betas = np.empty(p)
        gammas_singles = np.empty(p)
        gammas_pairs = np.empty(p)

        return cls(qaoa_circuit_params, betas, gammas_singles, gammas_pairs)

    def plot(self, ax=None, **kwargs):
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(self.betas, label="betas", marker="s", ls="", **kwargs)
        if not _is_iterable_empty(self.gammas_singles):
            ax.plot(self.gammas_singles,
                    label="gammas_singles", marker="^", ls="", **kwargs)
        if not _is_iterable_empty(self.gammas_pairs):
            ax.plot(self.gammas_pairs,
                    label="gammas_pairs", marker="v", ls="", **kwargs)
        ax.set_xlabel("timestep")
        # ax.grid(linestyle='--')
        ax.legend()



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABC
import numpy as np
from openqaoa.devices import DeviceLocal, DeviceBase

from openqaoa.rqaoa.rqaoa import custom_rqaoa
from openqaoa.problems.problem import QUBO
from openqaoa.problems.helper_functions import convert2serialize
from openqaoa.workflows.parameters.qaoa_parameters import CircuitProperties, BackendProperties, ClassicalOptimizer
from openqaoa.workflows.parameters.rqaoa_parameters import RqaoaParameters
from openqaoa.qaoa_parameters import Hamiltonian, QAOACircuitParams, create_qaoa_variational_params
from openqaoa.utilities import get_mixer_hamiltonian
from openqaoa.backends.qaoa_backend import get_qaoa_backend, DEVICE_NAME_TO_OBJECT_MAPPER, DEVICE_ACCESS_OBJECT_MAPPER
from openqaoa.optimizers.qaoa_optimizer import get_optimizer
from openqaoa.rqaoa import adaptive_rqaoa, custom_rqaoa


class Optimizer(ABC):
    """
    Abstract class to represent an optimizer
    """

    def asdict(self):
        pass


class QAOA(Optimizer):
    """
    A class implementing a QAOA workflow end to end.

    It's basic usage consists of 
    1. Initialization
    2. Compilation
    3. Optimization

    .. warning::
        To all our dear beta testers: the setter functions will most likely change. Bear with us as we figure our the smoother way to create the workflows :-)


    .. note::
        The attributes of the QAOA class should be initialized using the set methods of QAOA. For example, to set the circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
        circuit_properties: `CircuitProperties`
            The circuit properties of the QAOA workflow. Use to set depth `p`, choice of parametrisation, parameter initialisation strategies, mixer hamiltonians.
            For a complete list of its parameters and usage please see the method set_circuit_properties
        backend_properties: `BackendProperties`
            The backend properties of the QAOA workflow. Use to set the backend properties such as the number of shots and the cvar values.
            For a complete list of its parameters and usage please see the method set_backend_properties
        classical_optimizer: `ClassicalOptimizer`
            The classical optimiser properties of the QAOA workflow. Use to set the classical optimiser needed for the classical optimisation part of the QAOA routine.
            For a complete list of its parameters and usage please see the method set_classical_optimizer
        local_simulators: list[str]`
            A list containing the available local simulators
        mixer_hamil: Hamiltonian
            The desired mixer hamiltonian
        cost_hamil: Hamiltonian
            The desired mixer hamiltonian
        circuit_params: QAOACircuitParams
            the abstract and backend-agnostic representation of the underlying QAOA parameters
        variate_params: QAOAVariationalBaseParams
            The variational parameters. These are the parameters to be optimised by the classical optimiser
        backend: VQABaseBackend
            The openQAOA representation of the backend to be used to execute the quantum circuit
        optimizer: OptimizeVQA
            The classical optimiser
        results: `Result`
            Contains the logs of the optimisation process
        compiled: `Bool`
            A boolean flag to check whether the QAOA object has been correctly compiled at least once

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> q = QAOA()
    >>> q.compile(QUBO)
    >>> q.optimise()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    >>> q_custom = QAOA()
    >>> q_custom.set_circuit_properties(p=10, param_type='extended', init_type='ramp', mixer_hamiltonian='x')
    >>> q_custom.set_device_properties(device_location='qcs', device_name='Aspen-11', cloud_credentials={'name' : "Aspen11", 'as_qvm':True, 'execution_timeout' : 10, 'compiler_timeout':10})
    >>> q_custom.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> q_custom.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> q_custom.compile(qubo_problem)
    >>> q_custom.optimize()
    """

    def __init__(self, device=DeviceLocal('vectorized')):
        self.circuit_properties = CircuitProperties()
        self.device = device
        self.backend_properties = BackendProperties()
        self.classical_optimizer = ClassicalOptimizer()
        self.local_simulators = list(DEVICE_NAME_TO_OBJECT_MAPPER.keys())
        self.cloud_provider = list(DEVICE_ACCESS_OBJECT_MAPPER.keys())
        self.compiled = False

    def set_device(self, device: DeviceBase):
        """"
        Specify the device to be used by the QAOA.

        Parameters
        ----------
        location: `str`
            Can be either local, qcs, or ibmq
        name: `str`
            The name of the device to be used, for local simulators please refer to `q.local_simulators`.
            For cloud providers please refer to the provider's naming conventions
        """
        self.device = device

    def set_circuit_properties(self, **kwargs):
        """
        Specify the circuit properties to construct QAOA circuit

        Parameters
        -------------------
            qubit_register: `list`
                Select the desired qubits to run the QAOA program. Meant to be used as a qubit
                selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
            p: `int`
                Depth `p` of the QAOA circuit
            q: `int`
                Analogue of `p` of the QAOA circuit in the Fourier parameterisation
            param_type: `str`
                Choose the QAOA circuit parameterisation. Currently supported parameterisations include:
                `'standard'`: Standard QAOA parameterisation
                `'standard_w_bias'`: Standard QAOA parameterisation with a separate parameter for single-qubit terms.
                `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
                `'fourier'`: Fourier circuit parameterisation
                `'fourier_extended'`: Fourier circuit parameterisation with individual parameter for each qubit and term in Hamiltonian.
                `'fourier_w_bias'`: Fourier circuit parameterisation with aseparate parameter for single-qubit terms
            init_type: `str`
                Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
                `'rand'`: Randomly initialise circuit parameters
                `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit parameters (inspired from Quantum Annealing)
                `'custom'`: User specified initial circuit parameters
            mixer_hamiltonian: `str`
                Parameterisation of the mixer hamiltonian:
                `'x'`: Randomly initialise circuit parameters
                `'xy'`: Linear ramp from Hamiltonian initialisation of circuit 
            mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
                The connectivity of the qubits in the mixer Hamiltonian. Use only if `mixer_hamiltonian = xy`. The user can specify the 
                connectivity as a list of lists, a list of tuples, or a string chosen from ['full', 'chain', 'star'].
            mixer_coeffs: `list`
                The coefficients of the mixer Hamiltonian. By default all set to -1
            annealing_time: `float`
                Total time to run the QAOA program in the Annealing parameterisation (digitised annealing)
            linear_ramp_time: `float`
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            variational_params_dict: `dict`
                Dictionary object specifying the initial value of each circuit parameter for the chosen parameterisation, if the `init_type` is selected as `'custom'`.    
        """

        for key, value in kwargs.items():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError(
                    "Specified argument is not supported by the circuit")
        self.circuit_properties = CircuitProperties(**kwargs)

        return None

    def set_backend_properties(self, **kwargs):
        """
        Set the backend properties

        Parameters
        -------------------
            device: DeviceBase
            prepend_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            append_state: [Union[QuantumCircuitBase,List[complex], np.ndarray]
                The state prepended to the circuit.
            init_hadamard: bool
            Whether to apply a Hadamard gate to the beginning of the 
                QAOA part of the circuit.. Defaults to `True`
            n_shots: int
            Optional argument to specify the number of shots required to run QAOA computations
                on shot-based simulators and QPUs. Defaults to 100.
            cvar_alpha: float
                The value of alpha for the CVaR cost function
            qiskit_simulation_method: str, optional
                The method to be used for the simulation.
            noise_model: `qiskit.providers.aer.noise.NoiseModel`
                    The Qiskit noise model to be used for the simulation.
            active_reset:
                #TODO
            rewiring:
                Rewiring scheme to be used for Pyquil. 
                Either 'PRAGMA INITIAL_REWIRING "NAIVE"' or 
                'PRAGMA INITIAL_REWIRING "PARTIAL"'. If None, defaults to NAIVE
        """

        for key, value in kwargs.items():
            if hasattr(self.backend_properties, key):
                pass# setattr(self.backend_properties, key, value)
            else:
                raise ValueError(
                    f'Specified argument `{value}` for `{key}` in set_backend_properties is not supported')

        self.backend_properties = BackendProperties(**kwargs)
        return None

    def set_classical_optimizer(self, **kwargs):
        """
        Set the parameters for the classical optimizer to be used in the QAOA workflow

        Parameters
        -------------------
            method: str
                The classical optimization method. Choose from:
                ['imfil','bobyqa','snobfit']
                ['vgd', 'sgd', 'rmsprop'] 
                ['nelder-mead','powell','cg','bfgs','newton-cg','l-bfgs-b','cobyla'] 
            maxiter : Optional[int]
                Maximum number of iterations.
            maxfev : Optional[int]
                Maximum number of function evaluations.
            jac: str
                Method to compute the gradient vector. Choose from:
                ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']        
            hess: str
                Method to compute the hessian. Choose from:
                ['finite_difference', 'param_shift', 'stoch_param_shift', 'grad_spsa']
            constraints: scipy.optimize.LinearConstraints, scipy.optimize.NonlinearConstraints  
                Scipy-based constraints on parameters of optimization. Will be available soon
            bounds: scipy.optimize.Bounds
                Scipy-based bounds on parameters of optimization. Will be available soon
            tol : float
                Tolerance before the optimizer terminates; if `tol` is larger than
                the difference between two steps, terminate optimization.
            optimizer_options : dict
                Dictionary of optimiser-specific arguments.
                stepsize : float
                    Step size of each gradient descent step.
                decay : float
                    Stepsize decay parameter of RMSProp.
                eps : float
                    Small number to prevent division by zero for RMSProp.
                lambd : float
                    Small number to prevent singularity of QFIM matrix for Natural Gradient Descent.
            ramp_time: float
                The slope(rate) of linear ramp initialisation of QAOA parameters.
            jac_options : dict
                Dictionary that specifies gradient-computation options according to method chosen in 'jac'.
            hess_options : dict
                Dictionary that specifies Hessian-computation options according to method chosen in 'hess'.
        """
        for key, value in kwargs.items():
            if hasattr(self.classical_optimizer, key):
                pass #setattr(self.classical_optimizer, key, value)
            else:
                raise ValueError(
                    'Specified argument is not supported by the Classical Optimizer')

        self.classical_optimizer = ClassicalOptimizer(**kwargs)
        return None

    def compile(self, problem: QUBO = None, verbose: bool = False):
        """
        Initialise the trainable parameters for QAOA according to the specified
        strategies and by passing the problem statement

        .. note::
            Compilation is necessary because it is the moment where the problem statement and the QAOA instructions are used to build the actual QAOA circuit.

        .. tip::
            Set Verbose to false if you are running batch computations! 

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by QAOA
        verbose: bool
            Set True to have a summary of QAOA to displayed after compilation
        """

        assert isinstance(problem, QUBO), "The problem must be converted into QUBO form"
        
        self.cost_hamil = Hamiltonian.classical_hamiltonian(
            terms=problem.terms, coeffs=problem.weights, constant=problem.constant)
        
        self.mixer_hamil = get_mixer_hamiltonian(n_qubits=self.cost_hamil.n_qubits,
                                                 mixer_type=self.circuit_properties.mixer_hamiltonian,
                                                 qubit_connectivity=self.circuit_properties.mixer_qubit_connectivity,
                                                 coeffs=self.circuit_properties.mixer_coeffs)

        self.circuit_params = QAOACircuitParams(
            self.cost_hamil, self.mixer_hamil, p=self.circuit_properties.p)
        self.variate_params = create_qaoa_variational_params(qaoa_circuit_params=self.circuit_params,
                                                             params_type=self.circuit_properties.param_type,
                                                             init_type=self.circuit_properties.init_type, 
                                                             variational_params_dict=self.circuit_properties.variational_params_dict,
                                                             linear_ramp_time=self.circuit_properties.linear_ramp_time, 
                                                             q=self.circuit_properties.q, 
                                                             seed=self.circuit_properties.seed,
                                                             total_annealing_time=self.circuit_properties.annealing_time)

        self.backend = get_qaoa_backend(circuit_params=self.circuit_params,
                                        device=self.device,
                                        **self.backend_properties.__dict__)

        self.optimizer = get_optimizer(vqa_object=self.backend,
                                       variational_params=self.variate_params,
                                       optimizer_dict=self.classical_optimizer.asdict())

        self.compiled = True

        if verbose:
            print('\t \033[1m ### Summary ###\033[0m')
            print(f'OpenQAOA has been compiled with the following properties')
            print(
                f'Solving QAOA with \033[1m {self.device.device_name} \033[0m on  \033[1m{self.device.device_location}\033[0m')
            print(f'Using p={self.circuit_properties.p} with {self.circuit_properties.param_type} parameters initialized as {self.circuit_properties.init_type}')

            if self.device.device_name == 'vectorized':
                print(
                    f'OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations')

            else:
                print(
                    f'OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations. Each iteration will contain \033[1m{self.backend_properties.n_shots} shots\033[0m')
                print(
                    f'The total number of shots is set to maxiter*shots = {self.classical_optimizer.maxiter*self.backend_properties.n_shots}')

        return None

    def optimize(self, verbose=False):
        '''
        A method running the classical optimisation loop
        '''

        if self.compiled == False:
            raise ValueError('Please compile the QAOA before optimizing it!')

        self.optimizer.optimize()
        # TODO: results and qaoa_results will differ
        self.results = self.optimizer.qaoa_result

        if verbose:
            print(f'optimization completed.')
        return


class RQAOA(Optimizer):
    """
    RQAOA optimizer class.

    Attributes
    ----------
    algorithm: `str`
        A string contaning the name of the algorithm, here fixed to `rqaoa`
    qaoa: 'QAOA'
        QAOA class instance containing all the relevant information for the
        QAOA runs at each recursive step.
    rqaoa_parameters: 'RqaoaParameters'
        Set of parameters containing all the relevant information for the 
        recursive procedure.
    result: `dict`
        Dictionary containing the solution and all of the RQAOA procedure 
        information
    """

    def __init__(self, rqaoa_type: str = 'adaptive', qaoa: QAOA = QAOA()):
        """
        Initializes the RQAOA optimizer class.

        Parameters
        ----------
        rqaoa_type: `str`
            Recursive scheme for RQAOA algorithm. Choose from `custom` or `adaptive`
        q: `QAOA`
            QAOA instance specificying how QAOA is run within RQAOA
        """
        self.algorithm = 'rqaoa'
        self.qaoa = qaoa
        self.rqaoa_parameters = RqaoaParameters(rqaoa_type=rqaoa_type)
        self.rqaoa_mixer = {'type': self.qaoa.circuit_properties.mixer_hamiltonian,
                            'connectivity': self.qaoa.circuit_properties.mixer_qubit_connectivity,
                            'coeffs': self.qaoa.circuit_properties.mixer_coeffs}

    def set_rqaoa_parameters(self, **kwargs):
        """
        Sets the parameters for the RQAOA class.
        """
        for key, value in kwargs.items():
            if hasattr(self.rqaoa_parameters, key):
                setattr(self.rqaoa_parameters, key, value)
            else:
                raise ValueError(
                    f'Specified part {key},  {value} is not supported by RQAOA')

        return None

    def compile(self, problem: QUBO = None, verbose: bool = True):
        """
        Initialize the trainable parameters for QAOA according to the specified
        strategies and by passing the problem statement.

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by RQAOA
        verbose: bool
            !NotYetImplemented! Set true to have a summary of RQAOA to displayed after compilation
        """

        self.qaoa.compile(problem, verbose=False)

    def optimize(self):
        """
        Performs optimization using RQAOA with the `custom` method or the `adaptive` method.
        """
        if self.rqaoa_parameters.rqaoa_type == 'adaptive':
            res = adaptive_rqaoa(
                hamiltonian=self.qaoa.cost_hamil,
                mixer=self.rqaoa_mixer,
                p=self.qaoa.circuit_properties.p,
                n_max=self.rqaoa_parameters.n_max,
                n_cutoff=self.rqaoa_parameters.n_cutoff,
                device=self.qaoa.device,
                params_type=self.qaoa.circuit_properties.param_type,
                init_type=self.qaoa.circuit_properties.init_type,
                optimizer_dict=self.qaoa.classical_optimizer.asdict(),
                backend_properties=self.qaoa.backend_properties.__dict__)
        elif self.rqaoa_parameters.rqaoa_type == 'custom':
            res = custom_rqaoa(
                hamiltonian=self.qaoa.cost_hamil,
                mixer=self.rqaoa_mixer,
                p=self.qaoa.circuit_properties.p,
                n_cutoff=self.rqaoa_parameters.n_cutoff,
                steps=self.rqaoa_parameters.steps,
                device=self.qaoa.device,
                params_type=self.qaoa.circuit_properties.param_type,
                init_type=self.qaoa.circuit_properties.init_type,
                optimizer_dict=self.qaoa.classical_optimizer.asdict(),
                backend_properties=self.qaoa.backend_properties.__dict__)
        else:
            raise f'rqaoa_type {self.rqaoa_parameters.rqaoa_type} is not supported. Please selet either "adaptive" or "custom'

        self.result = res
        return None

    def asdict(self):
        attributes_dict = convert2serialize(self)
        return attributes_dict



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Collection of custom optimization algorithms to be used by Scipy `minimize`. Extends available scipy methods.

"""

import numpy as np
from scipy.optimize import OptimizeResult


def grad_descent(fun, x0, args=(), maxfev=None, stepsize=0.01,
                 maxiter=100, tol=10**(-6), jac=None, callback=None, **options):
    '''    
    Minimize a function `fun` using gradient descent. scipy.optimize.minimize compatible implementation of gradient descent for `method` == 'vgd'.

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    '''

    bestx = x0
    besty = fun(x0)
    funcalls = 1  # tracks no. of function evals.
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # compute gradient descent step
        testx = testx - stepsize*jac(testx)
        testy = np.real(fun(testx, *args))

        if np.abs(besty-testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(fun=besty, x=bestx, nit=niter,
                          nfev=funcalls, success=(niter > 1))


def rmsprop(fun, x0, args=(), maxfev=None, stepsize=0.01,
            maxiter=100, tol=10**(-6), jac=None, callback=None, decay=0.9, eps=1e-07, **options):
    '''    
    Minimize a function `fun` using RMSProp. scipy.optimize.minimize compatible implementation of RMSProp for `method` == 'rmsprop'.

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.
    decay : float
        Stepsize decay parameter of RMSProp.
    eps : float
        Small number to prevent division by zero.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    '''

    bestx = x0
    besty = fun(x0)

    sqgrad = jac(x0)**2
    funcalls = 1  # tracks no. of function evals
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # compute gradient descent step, scaled with adaptive RMS learning rate
        sqgrad = decay*sqgrad + (1-decay)*jac(testx)**2
        testx = testx - stepsize*jac(testx)/(np.sqrt(sqgrad) + eps)
        testy = np.real(fun(testx, *args))

        if np.abs(besty-testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(fun=besty, x=bestx, nit=niter,
                          nfev=funcalls, success=(niter > 1))


def newton_descent(fun, x0, args=(), maxfev=None, stepsize=0.01,
                   maxiter=100, tol=10**(-6), jac=None, hess=None, callback=None, **options):
    '''    
    Minimize a function `fun` using Newton's method, a second order method. 
    scipy.optimize.minimize compatible implementation of Newton's method for `method` == 'newton'.

    PARAMETERS
    ----------
    fun : callable
        Function to minimize
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    hess : callable
        Callable hessian function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    '''

    bestx = x0
    besty = fun(x0)
    funcalls = 1
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # Scale gradient with inverse of Hessian
        scaled_gradient = np.linalg.solve(hess(testx), jac(testx))

        # compute Newton descent step
        testx = testx - stepsize*scaled_gradient
        testy = np.real(fun(testx, *args))

        if np.abs(besty-testy) < tol:
            improved = False
        else:
            # if testy < besty:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(fun=besty, x=bestx, nit=niter,
                          nfev=funcalls, success=(niter > 1))


def natural_grad_descent(fun, x0, args=(), maxfev=None, stepsize=0.01,
                         maxiter=100, tol=10**(-6), lambd=0.001, jac=None, callback=None, **options):
    '''    
    Minimize a function `fun` using the natural gradient descent method.
    scipy.optimize.minimize compatible implementation of natural gradient descent for `method` == 'natural_grad_descent'.

    PARAMETERS
    ----------
    fun : QAOACost
        Function to minimize. Must be a `QAOACost` object, so that the method `qfim` can be called.
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    stepsize : float
        Step size of each gradient descent step.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    lambd : float
        Magnitude of the Identity regularization term to avoid singularity of the QFIM.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    '''

    bestx = x0
    besty = fun(x0)
    n_params = len(x0)
    funcalls = 1
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # scale gradient by left-multiplication with qfi matrix.
        qfim = options['qfim']
        scaled_gradient = np.linalg.solve(
            qfim(testx) + lambd*(np.identity(n_params)), jac(testx))

        # compute natural gradient descent step
        testx = testx - stepsize*scaled_gradient
        testy = np.real(fun(testx, *args))

        if np.abs(besty-testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(fun=besty, x=bestx, nit=niter,
                          nfev=funcalls, success=(niter > 1))


#### Experimental ############################################################

def SPSA(fun, x0, args=(), maxfev=None, a0=0.01, c0=0.01, A=1, alpha=0.602, gamma=0.101,
         maxiter=100, tol=10**(-6), jac=None, callback=None, **options):
    '''    
    Minimize a function `fun` using the Simultaneous Perturbation Stochastic Approximation (SPSA) method.
    scipy.optimize.minimize compatible implementation of SPSA for `method` == 'spsa'. Refer to https://www.jhuapl.edu/spsa/PDF-SPSA/Spall_Implementation_of_the_Simultaneous.PDF for in depth explanation of parameters and optimal parameter choices.

    PARAMETERS
    ----------
    fun : QAOACost
        Function to minimize. Must be a `QAOACost` object, so that the method `qfim` can be called.
    x0 : ndarray
        Initial guess.
    args : sequence, optional
        Arguments to pass to `func`.
    maxfev : int, optional
        Maximum number of function evaluations.
    a0 : float
        Initial stepsize.
    c0 : float
        Initial finite difference.
    A : float
        Initial decay constant. 
    alpha : float
        Decay rate parameter for `a`.
    gamma : float
        Decay rate parameter for `c`.
    maxiter : int, optional
        Maximum number of iterations.
    tol : float
        Tolerance before the optimizer terminates; if `tol` is larger than the difference between two steps, terminate optimization.
    jac : callable
        Callable gradient function.
    callback : callable, optional
        Called after each iteration, as ``callback(xk)``, where ``xk`` is the
        current parameter vector.

    RETURNS
    -------
    OptimizeResult : OptimizeResult
        Scipy OptimizeResult object.
    '''

    def grad_SPSA(params, c):
        delta = (2*np.random.randint(0, 2, size=len(params))-1)
        return np.real((fun(params + c*delta) - fun(params - c*delta))*delta/(2*c))

    bestx = x0
    besty = fun(x0)
    funcalls = 1  # tracks no. of function evals.
    niter = 0
    improved = True
    stop = False

    testx = np.copy(bestx)
    testy = np.real(fun(testx, *args))

    a = a0
    c = c0
    while improved and not stop and niter < maxiter:
        improved = False
        niter += 1

        # gain sequences
        a = a/(A+niter+1)**alpha
        c = c/(niter+1)**gamma

        # compute gradient descent step
        testx = testx - a*grad_SPSA(testx, c)
        testy = np.real(fun(testx, *args))

        if np.abs(besty-testy) < tol:
            improved = False

        else:
            besty = testy
            bestx = testx
            improved = True

        if callback is not None:
            callback(bestx)
        if maxfev is not None and funcalls >= maxfev:
            stop = True
            break

    return OptimizeResult(fun=besty, x=bestx, nit=niter,
                          nfev=funcalls, success=(niter > 1))


'''
def stochastic_grad_descent(fun, x0, jac, args=(), stepsize=0.001, mass=0.9, startiter=0, maxiter=1000, callback=None, **options ):
    """
    Scipy "OptimizeResult" object for method == 'sgd'
    scipy.optimize.minimize compatible implementation of stochastic gradient descent with momentum.
    Adapted from ``autograd/misc/optimizers.py``.

    """
    x = x0
    velocity = np.zeros_like(x)
    print("SGD Test")

    for i in range(startiter, startiter + maxiter):
        g = jac(x)

        velocity = mass * velocity - (1.0 - mass) * g
        x = x + stepsize * velocity

        if callback is not None:
            callback(x)

    return OptimizeResult(x=x, fun=fun(x), jac=g, nit=i+1, nfev=i+1, success=True)
'''



#   Copyright 2021 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Utility and convenience functions for a number of QAOA applications.
"""

from typing import Optional, Union, List, Tuple
import itertools
import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
from .qaoa_parameters import Hamiltonian, PauliOp, QAOAVariationalBaseParams


def X_mixer_hamiltonian(n_qubits: int,
                        coeffs: List[float] = None) -> Hamiltonian:
    """Construct a Hamiltonian object to implement the X mixer.

    Parameters
    ----------
    n_qubits: `int`
        The number of qubits in the mixer Hamiltonian.
    coeffs: `list`, optional
        The coefficients of the X terms in the Hamiltonian.

    Returns
    -------
    hamiltonian: `Hamiltonian`
        The Hamiltonian object corresponding to the X mixer.
    """
    # If no coefficients provided, set all to -1
    coeffs = [-1]*n_qubits if coeffs is None else coeffs

    # Initialize list of terms
    terms = []

    # Generate terms in the X mixer
    for i in range(n_qubits):
        terms.append(PauliOp.X(i))
    
    # Define mixer Hamiltonian
    hamiltonian = Hamiltonian(pauli_terms=terms, coeffs=coeffs, constant=0)

    return hamiltonian


def XY_mixer_hamiltonian(n_qubits: int,
                         qubit_connectivity: Union[List[list],
                                                   List[tuple], str] = 'full',
                         coeffs: List[float] = None) -> Hamiltonian:
    """
    Construct a Hamiltonian object to implement the XY mixer.

    Parameters
    ----------
    n_qubits: `int`
        The number of qubits in the system.
    qubit_connectivity: `Union[List[list],List[tuple], str]`, optional
        The connectivity of the qubits in the mixer Hamiltonian.
    coeffs: `list`, optional
        The coefficients of the XY terms in the Hamiltonian.

    Returns
    -------
    hamiltonian: Hamiltonian
        The Hamiltonian object corresponding to the XY mixer.

    Notes
    -----
        The XY mixer is not implemented with $$RXY$$ Gates, but with $$H_{XY} = \\frac{1}{2}(\\sum_{i,j} X_iX_j+Y_iY_j)$$
    """
    # Set of topologies supported by default
    connectivity_topology_dict = {'full': list(itertools.combinations(range(n_qubits), 2)),
                                  'chain': [(i, i+1) for i in range(n_qubits-1)],
                                  'star': [(0, i+1) for i in range(n_qubits-1)]}

    # Check if input connectivity is a default value
    if isinstance(qubit_connectivity, str):
        try:
            # Define qubit connectivity from default value
            qubit_connectivity = connectivity_topology_dict[qubit_connectivity]
        except KeyError:
            raise ValueError(
                f'Please choose connection topology from {list(connectivity_topology_dict.keys())}')

    # Define connectivty according to user input
    else:
        # Extract indices from connectivity
        indices = set([qubit for term in qubit_connectivity for qubit in term])

        # Ensure all indices are defined within the range of number of qubits
        assert max(indices) <= n_qubits - \
            1, 'Qubit index in connectivity list is out of range'
        assert min(indices) >= 0, 'Qubit index should be a positive integer'

    # If no coefficients provided, set all to the number of terms
    coeffs = [0.5]*2*len(qubit_connectivity) if coeffs is None else coeffs

    # Initialize list of terms
    terms = []

    # Generate terms in the XY mixer
    for pair in qubit_connectivity:
        i, j = pair
        terms.append(PauliOp.X(i)@PauliOp.X(j))
        terms.append(PauliOp.Y(i)@PauliOp.Y(j))

    # Define mixer Hamiltonian
    hamiltonian = Hamiltonian(pauli_terms=terms, coeffs=coeffs, constant=0)

    return hamiltonian


def get_mixer_hamiltonian(n_qubits: int, mixer_type: str = 'x', qubit_connectivity: Union[List[list],List[tuple], str] = None, coeffs: List[float] = None):
    """
    Parameters
    ----------
    n_qubits: `int`
        Number of qubits in the Hamiltonian.
    mixer_type: `str`, optional
        Name of the mixer Hamiltonian. Choose from `x` or `xy`.
    qubit_connectivity: `list` or `str`, optional
        The connectivity of the qubits in the mixer Hamiltonian.
    coeffs: `list`, optional
        The coefficients of the terms in the Hamiltonian.

    Returns:
    --------
    mixer: `Hamiltonian`
        Hamiltonian object containing the specificied mixer.
    """

    # Return mixer Hamiltonian according to specified type
    if mixer_type == 'x':
        mixer = X_mixer_hamiltonian(n_qubits, coeffs)
    else:
        mixer = XY_mixer_hamiltonian(n_qubits, qubit_connectivity, coeffs)

    return mixer


################################################################################
# METHODS FOR PRINTING HAMILTONIANS AND GRAPHS, AND PRINTING ONE FROM EACH OTHER
################################################################################


def graph_from_hamiltonian(hamiltonian: Hamiltonian) -> nx.Graph:
    """
    Creates a networkx graph corresponding to a specified problem Hamiltonian.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        The Hamiltonian of interest. Must be specified a Hamiltonian object.

    Returns
    -------
    G: `Networkx Graph`
        The corresponding networkx graph with the edge weights being the
        two-qubit coupling coefficients,
        and the node weights being the single-qubit bias terms.

    Notes
    -----
    This function cannot handle non-QUBO terms.
    Linear terms are stored as nodes with weights.
    """
    # Define graph
    G = nx.Graph()

    # Add nodes for each qubit in the register
    for qubit in hamiltonian.qureg:
        G.add_node(qubit, weight=0)

    # Add each term to the graph as an attribute
    for term, weight in zip(hamiltonian.terms, hamiltonian.coeffs):

        # Extract indices from Pauli term
        term_tuple = term.qubit_indices

        # If term is linear add as a node with a weight attribute
        if(len(term) == 1):
            G.add_node(term_tuple[0], weight=weight)

        # If term is quadratic add as an edge with a weight attribute
        elif(len(term) == 2):
            G.add_edge(term_tuple[0], term_tuple[1], weight=weight)

    return G


def hamiltonian_from_graph(G: nx.Graph) -> Hamiltonian:
    """
    Builds a cost Hamiltonian as a collection of PauliOp objects
    from a specified networkx graph, extracting any node and edge weights.

    Parameters
    ----------
    G: `Networkx Graph`
        The specified networkx graph.

    Returns
    -------
    hamiltonian: `Hamiltonian`
        The Hamiltonian object constructed from the specified graph.
    """
    # Node bias terms
    nodes_info = nx.get_node_attributes(G, 'weight')
    singlet_terms = [(node,) for node in nodes_info.keys()]
    singlet_coeffs = list(nodes_info.values())

    # Edge terms
    edges_info = nx.get_edge_attributes(G, 'weight')
    pair_terms = list(edges_info.keys())
    pair_coeffs = list(edges_info.values())

    # Collect all terms and coefficients
    terms = singlet_terms + pair_terms
    coeffs = singlet_coeffs + pair_coeffs
    
    # Define Hamiltonian
    hamiltonian = Hamiltonian.classical_hamiltonian(
        terms=terms, coeffs=coeffs, constant=0)

    return hamiltonian


def random_k_regular_graph(degree: int,
                           nodes: List[int],
                           seed: int = None,
                           weighted: bool = False,
                           biases: bool = False) -> nx.Graph:
    """
    Produces a random graph with specified number of nodes, each having degree k.

    Parameters
    ----------
    degree: `int`
        Desired degree for the nodes.
    nodes: `list`
        The node set of the graph.
    seed: `int`, optional
        A seed for the random number generator.
    weighted: `bool`, optional
        Whether the edge weights should be uniform or different. If false, all weights are set to 1.
        If true, the weight is set to a random number drawn from the uniform distribution in the
        interval 0 to 1.
    biases: `bool`, optional
        Whether or not the graph nodes should be assigned a weight.
        If true, the weight is set to a random number drawn from the uniform
        distribution in the interval 0 to 1.

    Returns
    -------
    nx.Graph: `Networkx Graph`
        A graph with the properties as specified.
    """
    # Set numpy seed
    np.random.seed(seed=seed)

    # Create a random regular graph on the nodes
    G = nx.random_regular_graph(degree, len(nodes), seed)

    # Relabel nodes
    nx.relabel_nodes(G, {i: n for i, n in enumerate(nodes)})

    # Add edges between nodes
    for edge in G.edges():

        # If weighted attribute is False, all weights are set to 1
        if not weighted:
            G[edge[0]][edge[1]]['weight'] = 1

        # If weighted attribute is True, weights are assigned as random integers
        else:
            G[edge[0]][edge[1]]['weight'] = np.random.rand()

    # If biases attribute is True, add node weights as random integers
    if biases:
        for node in G.nodes():
            G.nodes[node]['weight'] = np.random.rand()

    return G

def plot_graph(G: nx.Graph, ax=None, colormap='seismic') -> None:
    """
    Plots a networkx graph.

    Parameters
    ----------
    G: `Networkx Graph`
        The networkx graph of interest.
    ax: `Matplotlib axes object`, optional
        Matplotlib axes to plot on. Defaults to None.
    colormap: `str`, optional
        Colormap to use for plotting. Defaults to 'seismic'.
    """
    
    # Create plot figure
    fig = plt.figure(figsize=(10, 6))
        
    # Extract all graph attributes
    biases_and_nodes = nx.get_node_attributes(G, 'weight')
    biases = list(biases_and_nodes.values())
    edges_and_weights = nx.get_edge_attributes(G, 'weight')
    pos = nx.shell_layout(G)

    # extract minimum and maximum weights for side bar limits
    weights = list(edges_and_weights.values())
    # Define color map
    cmap = plt.cm.get_cmap(colormap)
    
    if len(set(weights)) > 1:
        edge_vmin = min(weights)
        edge_vmax = max(weights)


        # Define normalized color map
        sm = plt.cm.ScalarMappable(cmap=cmap,
                                norm=plt.Normalize(vmin=edge_vmin, vmax=edge_vmax))
        # Add colormap to plot
        cbar = plt.colorbar(sm,pad=0.08)
        cbar.ax.set_ylabel('Edge Weights', rotation=270,labelpad=15)
    else:
        weights = [1] * len(G.edges())
        edge_vmin = None
        edge_vmax = None
        cmap = None    
    
    # If biases are present define reference values and color map for side bar
    if len(set(biases)) > 1:
        cmap = plt.cm.get_cmap(colormap)
        vmin = min(biases)
        vmax = max(biases)
        sm2 = plt.cm.ScalarMappable(cmap=cmap,
                                    norm=plt.Normalize(vmin=vmin, vmax=vmax))
        cbar2 = plt.colorbar(sm2, location='left')
        cbar2.ax.set_ylabel('Single Qubit Biases', rotation=90)

        # Draw graph
        nx.draw(G, pos, node_size=500, node_color=biases, edge_color=weights, width=2.5, cmap=cmap,
                edge_cmap=cmap, vmin=vmin, vmax=vmax, edge_vmin=edge_vmin,
                edge_vmax=edge_vmax, with_labels=True)

    else:
        # Draw graph
        nx.draw(G, pos, node_size=500, edge_color=weights, width=2.5,
                edge_cmap=cmap, edge_vmin=edge_vmin,
                edge_vmax=edge_vmax, with_labels=True)
    
    # Show plot
    
    plt.show
    return None

def random_classical_hamiltonian(reg: List[int],
                                 seed: int = None,
                                 weighted: bool = True,
                                 biases: bool = True,
                                 constant: int = 0) -> Hamiltonian:
    """
    Creates a random classical cost hamiltonian.

    Parameters
    ----------
    reg: `list`
        Register to build the hamiltonian on.
    seed: `int`, optional
        A seed for the random number generator. Defaults to None.
    weighted: `bool`, optional
        Whether the edge weights should be uniform or different. If false, all
        weights are set to 1. If true, the weight is set to a random number 
        drawn from the uniform distribution in the interval 0 to 1. Defaults to
        True.
    biases: `bool`, optional
        Whether or not the graph nodes should be assigned a weight.
        If true, the weight is set to a random number drawn from the uniform
        distribution in the interval 0 to 1. Defaults to True.
    constant: `int`, optional
        The constant term in the Hamiltonian. Defaults to 0.

    Returns
    -------
    random_hamil: `Hamiltonian`
        A random hamiltonian with randomly selected terms and coefficients and 
        with the specified constant term.

    Notes
    -----
    Randomly selects which qubits that will have a bias term, then assigns them a bias coefficient.
    Randomly selects which qubit pairs will have a coupling term, then assigns them a coupling coefficient.
    In both cases, the random coefficient is drawn from the uniform distribution on the interval [0,1).
    """
    # Set the random seed
    np.random.seed(seed=seed)

    # Initialize terms and weights
    terms = []
    weights = []

    # If biases attribute is True, add lineat terms
    if biases:
        # Choose a random set of qubits to add linear terms
        n_biases = np.random.randint(len(reg))
        bias_qubits = np.random.choice(reg, n_biases)

        # Generate coefficients for linear terms
        bias_coeffs = np.random.rand(
            n_biases) if weighted else np.ones(n_biases)

        # Store linear terms and coefficients
        for qubit, coeff in zip(bias_qubits, bias_coeffs):
            terms.append([qubit])
            weights.append(coeff)
    
    # Generate quiadratic terms, scanning all possible combinations
    for q1, q2 in itertools.combinations(reg, 2):

        # Choose at random to couple terms
        are_coupled = np.random.randint(2)

        # If coupled, generate coefficients and store along with term
        if are_coupled:
            couple_coeff = np.random.rand() if weighted else 1
            terms.append([q1, q2])
            weights.append(couple_coeff)

    # Ensure each term has an associated weight
    assert len(terms) == len(
        weights), "Each term should have an associated weight"

    # Define classical Hamiltonian
    hamiltonian = Hamiltonian.classical_hamiltonian(
        terms, weights, constant=constant)

    return hamiltonian


################################################################################
# HAMILTONIANS AND DATA
################################################################################


def ground_state_hamiltonian(hamiltonian: Hamiltonian, bounded = True) -> Tuple[float, list]:
    """
    Computes the exact ground state and ground state energy of a classical Hamiltonian. Uses standard numpy module.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose ground state properties are computed.
    bounded: `bool`, optional
        If set to True, the function will not perform computations for qubit
        numbers above 25. If False, the user can specify any number. Defaults
        to True.

    Returns
    -------
    min_energy: `float`
        The minimum eigenvalue of the cost Hamiltonian.

    config: `np.array`
        The minimum energy eigenvector as a binary array
        configuration: qubit-0 as the first element in the sequence.
    """
    # Extract number of qubits 
    n_qubits = hamiltonian.n_qubits
    
    # If number of qubits is too high warn the user
    if bounded and n_qubits > 25:
        raise ValueError("The number of qubits is too high, computation could take a long time. If still want to proceed set argument `bounded` to False")

    # Generate qubit register
    register = range(n_qubits)

    # Intialize energies
    energies = np.zeros(2**len(register))

    # Obtain spectrum, scanning term by term
    for i, term in enumerate(hamiltonian.terms):

        # Extract coefficient
        out = np.real(hamiltonian.coeffs[i])

        # Compute tensor product
        for qubit in register:
            if qubit in term.qubit_indices:
                out = np.kron([1, -1], out)
            else:
                out = np.kron([1, 1], out)

        # Add energy contribution of the term
        energies += out
    
    # Add constant term to the spectrum
    energies += hamiltonian.constant

    # Extract minimum energy
    min_energy = np.min(energies)

    # Extract indices of minimum energies
    indices = np.where(energies == min_energy)[0]

    # Generate ground states
    config_strings = [np.binary_repr(index, len(register))[::-1]
                      for index in indices]

    return min_energy, config_strings


def bitstring_energy(hamiltonian: Hamiltonian,
                     bitstring: Union[List[int], str]) -> float:
    """
    Computes the energy of a given bitstring with respect to a classical cost Hamiltonian.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object determining the energy levels.
    bitstring : `list` or `str`
        A list of integers 0 and 1, or a string, representing a configuration.

    Returns
    -------
    energy: `float`
        The energy of the given bitstring with respect to the cost Hamiltonian.
    """
    # Initialize energy value
    energy = 0
    
    # Compute energy contribution term by term
    for i, term in enumerate(hamiltonian.terms):

        # Compute sign of spin interaction term
        variables_product = np.prod([(-1)**int(bitstring[k]) for k in term.qubit_indices])

        # Add energy contribution
        energy += hamiltonian.coeffs[i]*variables_product
    
    # Add constant contribution
    energy += hamiltonian.constant

    return energy


def energy_expectation(hamiltonian: Hamiltonian,
                       measurement_counts: dict) -> float:
    """
    Computes the energy expectation value from a set of measurement counts, with respect to a classical cost Hamiltonian.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object determining the energy levels.
    measurement_counts : `dict`
        Measurement counts dictionary for which to calculate energy expectation.

    Returns
    -------
    energy: `float`
        The energy expectation value for the set of measurement outcomes
    """

    # Starting value for the energy
    energy = 0

    # Number of measurement shots
    shots = sum(measurement_counts.values())

    # Compute average energy adding one by one the contribution from each state
    for state, prob in measurement_counts.items():

        # Number of ones (spins pointing down) from the specific configuration for each Hamiltonian term
        num_ones_list = [sum([int(state[i]) for i in term.qubit_indices])
                         for term in hamiltonian.terms]

        # Compute configuration energy
        config_energy = sum([hamiltonian.coeffs[i] if num_ones % 2 == 0 else -1*hamiltonian.coeffs[i]
                            for i, num_ones in enumerate(num_ones_list)])

        # Add contribution to total energy
        energy += prob*config_energy

    # Normalize with respect to the number of shots
    energy *= (1/shots)

    return energy


def energy_spectrum_hamiltonian(hamiltonian: Hamiltonian) -> np.ndarray:
    """
    Computes exactly the energy spectrum of the hamiltonian defined by terms
    and weights and its corresponding configuration of variables. Uses 
    standard numpy module.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose spectrum is computed.

    Returns
    -------
    energies: `np.ndarray`
        The energy spectra of the given hamiltonian
    """
    # Extract number of qubits
    n_qubits = hamiltonian.n_qubits

    # Define qubit register
    register = range(n_qubits)

    # Intialize energies
    energies = np.zeros((2**len(register)))

    # Obtain spectrum, scanning term by term
    for i, term in enumerate(hamiltonian.terms):

        # Extract coefficients
        out = np.real(hamiltonian.coeffs[i])

        # Compute tensor product 
        for qubit in register:
            if qubit in term.qubit_indices:
                out = np.kron([1, -1], out)
            else:
                out = np.kron([1, 1], out)

        # Add energy contribution of the term
        energies += out

    # Add constant term to the spectrum
    energies = energies + hamiltonian.constant

    return energies


def plot_energy_spectrum(hamiltonian: Hamiltonian,
                         high_k_states: Optional[int] = None,
                         low_k_states: Optional[int] = None,
                         ax=None,
                         cmap='winter') -> None:
    """
    Compute and plot the energy spectrum of a given hamiltonian on
    a matplotlib figure.

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Hamiltonian object whose spectrum is computed.
    high_k_states: `int`, optional
        Optionally plot the highest k energy levels. Defaults to None.
    low_k_states: `int`, optional
        Optionally, plot the lowest k energy levels. Defaults to None.
    ax: Matplotlib axes object, optional
        Axes to plot on. Defaults to None.
    cmap: `str`, optional
        Specify the matplotlib colormap to use for plotting. Defaults to 'winter'.
    """
    # Compute energy spectrum
    energies = energy_spectrum_hamiltonian(hamiltonian)

    # Extract energy levels and their degeneracy
    unique_energies, degeneracy = np.unique(energies, return_counts=True)

    # If required extract highest or lowest k energy levels
    if high_k_states is not None:
        unique_energies, degeneracy = unique_energies[-high_k_states:], degeneracy[-high_k_states:]
    elif low_k_states is not None:
        unique_energies, degeneracy = unique_energies[:low_k_states], degeneracy[:low_k_states]

    # Define colormap
    cmap = plt.cm.get_cmap(cmap, len(unique_energies))

    # If no axis provided, define figure
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(3, 5))

    # Plot energy levels
    for i, energy in enumerate(unique_energies):
        ax.axhline(energy, label=f'Degeneracy={degeneracy[i]}', color=cmap(i))

    # Set axis attributes and legend
    ax.set(xticks=[], yticks=unique_energies, ylabel='Energy[a.u.]',
           title='Hamiltonian Energy spectrum')
    ax.legend(loc='center left', fontsize=8)

    return None


def low_energy_states(hamiltonian: Hamiltonian,
                      threshold_per: float) -> Tuple[float, list]:
    """
    Return threshold energy and the low energy states of the
    specified hamiltonian which are below this threshold. 

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Compute the low energy states of this Hamiltonian
    threshold_per: `float`
        Threshold percentage away from the ground state, defining the energy we window we search
        in for low energy states.

    Returns
    -------
    low_energy_threshold: `float`
        The energy threshold below which we retrieve states. 
    states: `list`
        The list of low energy states that lie below the low
        energy threshold.

    Notes
    -----
    The threshold is calculated as `threshols_per` factors away from the
    ground state of the Hamiltonian.
    """
    # Asserr threshold is bounded between 0 and 1
    assert threshold_per >= 0.0, "Threshold percentage should be above 0"
    assert threshold_per <= 1.0, "Threshold percentage should be below 1"

    # Compute energy spectrum of the Hamiltonian
    energies = energy_spectrum_hamiltonian(hamiltonian)

    # Extract ground state and highest excited state
    ground_state_energy = np.min(energies)
    highest_state_energy = np.max(energies)

    # Compute the low energy threshols
    low_energy_threshold = ground_state_energy + threshold_per * \
        np.abs(highest_state_energy-ground_state_energy)

    # Initilize inndices for low energy states
    low_energy_indices = []

    # Obtain indices for energies below the threshold
    for idx in range(len(energies)):
        if energies[idx] <= low_energy_threshold:
            low_energy_indices.append(idx)
    
    # Extract states from the Hamiltonian spectrum
    states = [np.binary_repr(index, hamiltonian.n_qubits)[::-1]
              for index in low_energy_indices]

    return low_energy_threshold, states


def low_energy_states_overlap(hamiltonian: Hamiltonian,
                              threshold_per: float,
                              prob_dict: dict) -> float:
    """
    Calculates the overlap between the low energy states of a Hamiltonian,
    below a specific threshold away from the ground state energy, and an input
    state, expressed in terms of a probability dictionary. 

    Parameters
    ----------
    hamiltonian: `Hamiltonian`
        Compute overlap with respect to energies of this Hamiltonian
    threshold_per: `float`
        Threshold percentage away from the ground state, defining the energy we window we search
        in for low energy states.
    prob_dict: `Dict`
        The measurement outcome dictionary generated from the 
        circuit execution.

    Returns
    -------
    total_overlap: `float`
        The total overlap with the low-energy states.

    Notes
    -----
    The threshold is calculated as `threshold_per` factors away from the ground state of the Hamiltonain.
    For `threshold_per=0` the function returns the ground state overlap of the QAOA output.
    """
    # Extract number of qubits from probability dictionary
    n_qubits = len(list(prob_dict.keys())[0])

    # Ensure number of qubits matches the number of qubits registered in the Hamiltonian
    assert n_qubits == hamiltonian.n_qubits, "Number of qubits in the Hamiltonian does not match the probabilities specified"

    # Extract low energy states
    _, states = low_energy_states(hamiltonian, threshold_per)

    # Compute overlap
    total_overlap = sum([prob_dict[state]
                        for state in states])/sum(list(prob_dict.values()))

    return total_overlap


def exp_val_single(spin: int, prob_dict: dict):
    """
    Computes expectation value <Z> of a given spin.

    Parameters
    ----------
    spin: `int`
        Spin whose expectation value we compute.
    prob_dict: `dict`
        Dictionary containing the configuration probabilities of each spin.

    Returns
    -------
    exp_val: `float`
        Expectation value of the spin
    """

    # Initialize expectation value
    exp_val = 0

    # Compute correlation
    for bitstring, prob in prob_dict.items():

        # If 0, spin is pointing up, else it is pointing down
        Z = int(bitstring[spin])

        # Add contribution if spin points up or subtract if points down
        exp_val += -prob if Z > 0 else prob

    return exp_val


def exp_val_pair(spins: tuple, prob_dict: dict):
    """
    Computes the correlation Mij = <Z_{i}Z_{j}> between qubits i,j using the QAOA optimized 
    wavefunction.

    NOTE: In the presence of linear terms the <Z_{i}><Z_{j}> contribution needs to be
    subtracted later. This is done in the exp_val_hamiltonian_termwise() function used as a 
    wrapper for this function. 

    Parameters
    ----------
    spins: `tuple`
        Tuple containing the spins whose correlation is computed.
    prob_dict: `dict`
        The dictionary containing the configuration probabilities of each spin.

    Returns
    -------
    corr: `float`
        Correlation between the two spins in the term.

    """

    # Initialize correlation
    corr = 0

    # Compute correlation
    for bitstring, prob in prob_dict.items():

        # If 0 or 2, spins are aligned, else they are anti-aligned
        num_ones = sum([int(bitstring[i]) for i in spins])

        # Add contribution if spins aligned or subtract if anti-aligned
        corr += prob if num_ones % 2 == 0 else -prob

    return corr


def exp_val_hamiltonian_termwise(variational_params: QAOAVariationalBaseParams,
                                qaoa_backend,
                                hamiltonian: Hamiltonian, 
                                mixer_type:str, 
                                p: int,
                                qaoa_optimized_angles: Optional[list] = None,
                                qaoa_optimized_counts: Optional[dict] = None,
                                analytical: bool = True):
    """
    Computes the single spin expectation values <Z_{i}> and the correlation matrix Mij = <Z_{i}Z_{j}>,
    using the optimization results obtained from QAOA tranining the specified QAOA cost backend.

    Parameters
    ----------
    variational_params: `QAOAVariationalBaseParams`
        Set of variational parameters in the QAOA ansatz.
    qaoa_backend: `QAOABaseBackend`
        Chosen backend on which QAOA is performed.
    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.
    p: `int`
        Number of layers in QAOA ansatz.
    qaoa_optimized_angles: `list`
        Optimized angles of the underlying QAOA.
    qaoa_optimized_counts: `dict`
        Dictionary containing the measurement counts of optimized QAOA circuit.
    analytical: `bool`
        Boolean that indicates whether to use analytical or numerical expectation
        calculation methods.

    Returns
    -------
    exp_vals_z: `np.array`
        Single spin expectation values as a numpy array.
    corr_matrix: `np.array`
        Correlation matrix as a numpy Matrix object.
    """

    # Define number of qubits, problem hamiltonian and QAOA parameters
    n_qubits = hamiltonian.n_qubits

    # Extract Hamiltonian terms
    terms = list(hamiltonian.terms)
    
    # Initialize the z expectation values and correlation matrix with 0s
    exp_vals_z = np.zeros(n_qubits)
    corr_matrix = np.zeros((n_qubits, n_qubits))

    # If single layer ansatz use analytical results
    if (analytical == True and p == 1 and mixer_type == 'x' and
        isinstance(qaoa_optimized_angles, list)):

        # Compute expectation values and correlations of terms present in the Hamiltonian
        for term in terms:

            # If bias term compute expectation value
            if len(term) == 1:
                i = term.qubit_indices[0]
                exp_vals_z[i] = exp_val_single_analytical(
                    i, hamiltonian, qaoa_optimized_angles)

            # If two-body term compute correlation
            elif len(term) == 2:
                i, j = term.qubit_indices
                corr_matrix[i][j] = exp_val_pair_analytical(
                    (i, j), hamiltonian, qaoa_optimized_angles)

            # If constant term, ignore
            else:
                continue

    # If multilayer ansatz, perform numerical computation
    else:

        if isinstance(qaoa_optimized_counts, dict):
            counts_dict = qaoa_optimized_counts
        else:
            raise ValueError("Please specify optimized counts to compute expectation values.")

        # Compute expectation values and correlations of terms present in the Hamiltonian
        for term in terms:

            # If bias term compute expectation value
            if len(term) == 1:
                i = term.qubit_indices[0]
                exp_vals_z[i] = exp_val_single(i, counts_dict)

            # If two-body term compute correlation
            elif len(term) == 2:
                i, j = term.qubit_indices
                corr_matrix[i][j] = exp_val_pair((i, j), counts_dict)

            # If constant term, ignore
            if len(term) == 0:
                continue

    # Remove expectation value contribution from the correlations
    corr_matrix -= np.outer(exp_vals_z, exp_vals_z)

    return exp_vals_z, corr_matrix


################################################################################
# ANALYTIC & KNOWN FORMULAE
################################################################################


def exp_val_single_analytical(spin: int, hamiltonian: Hamiltonian, qaoa_angles: tuple):
    """
    Computes the single spin expectation value <Z> from an analytically
    derived expression for a single layer QAOA Ansatz. 
    
    NOTE: Only valid for single layer QAOA Ansatz with X mixer Hamiltonian.

    Parameters
    ----------
    spin: `int`
        The spin whose expectation value we compute.

    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.

    qaoa_angles: `tuple`
        Pair of (gamma,beta) angles defined from QAOA ansatz and
        obtained in the QAOA process.

    Returns:
    -------
    exp_val: `float`
        Spin expectation value <Z>.
    """

    # Number of qubits in the system
    n_qubits = hamiltonian.n_qubits

    # Extract graph properties of the Hamiltonian
    terms = list(hamiltonian.terms)
    edges = [terms[j].qubit_indices for j in range(len(terms))]
    weights = hamiltonian.coeffs

    # Hamiltonian from graph definitions
    hamil_graph = dict(zip(edges, weights))

    # Spin biases
    h_u = hamil_graph[(spin,)] if hamil_graph.get((spin,)) is not None else 0

    # QAOA angles
    beta, gamma = qaoa_angles

    # Spin register as a list without the spin we focus on
    iter_qubits = [j for j in range(0, spin)] + \
        [j for j in range(spin+1, n_qubits)]

    # Initialize products
    exp_val = -np.sin(2*beta) * np.sin(2*gamma*h_u)

    # Loop over edges connecting u and v to other spins
    for n in iter_qubits:

        # Edges between the spin and others in the register
        edge = tuple([min(spin, n), max(spin, n)])

        # If edge not present in the graph the associated weight is set to 0
        J_un = 0 if hamil_graph.get(edge) is None else hamil_graph[edge]

        # Add factor to the products
        exp_val *= np.cos(2*gamma*J_un)

    return exp_val


def exp_val_pair_analytical(spins: tuple, hamiltonian: Hamiltonian, qaoa_angles: tuple):
    """
    Computes <Z_{i}Z_{j}> correlation between apair of spins analytically. It is an extension from the 
    expression derived by Bravyi et al. in arXiv:1910.08980 which includes the effect of biases. 

    NOTE: Only valid for single layer QAOA Ansatz with X mixer Hamiltonian.

    NOTE: In the presence of linear terms the <Z_{i}><Z_{j}> contribution needs to be
    subtracted later. This is done in the exp_val_hamiltonian_termwise() function used as a 
    wrapper for this function. 

    NOTE: OpenQAOA uses a different sign convention for the QAOA Ansatz than Bravy et al. - there is 
    a relative minus sign between the cost function and the mixer in OpenQAOA, which 
    is accounted for in this implementation. Additionally, the result below is valid 
    for a Hadamard state initialization and in the absence of bias terms in the Hamiltonian.

    Parameters
    ----------
    spins: `tuple`
        Pair of spins whose correlation we compute.

    hamiltonian: `Hamiltonian`
        Hamiltonian object containing the problem statement.

    qaoa_angles: `tuple`
        Pair of (gamma,beta) angles defined from QAOA ansatz and
        obtained in the QAOA process.

    Returns
    -------
    corr:
        Correlation <ZZ> between the specified spin pair.
    """

    # Number of qubits in the system
    n_qubits = hamiltonian.n_qubits

    # Extract graph properties of the Hamiltonian
    terms = list(hamiltonian.terms)
    edges = [terms[j].qubit_indices for j in range(len(terms))]
    weights = hamiltonian.coeffs

    # Hamiltonian from graph definitions
    hamil_graph = dict(zip(edges, weights))

    # Spins whose correlation we compute
    u, v = spins

    # Coupling between the spins
    J_uv = hamil_graph[spins] if hamil_graph.get(spins) is not None else 0

    # Spin biases
    h_u = hamil_graph[(u,)] if hamil_graph.get((u,)) is not None else 0
    h_v = hamil_graph[(v,)] if hamil_graph.get((v,)) is not None else 0

    # QAOA angles
    beta, gamma = qaoa_angles

    # Factors in the expression
    s = np.sin(2*beta)
    c = np.cos(2*beta)

    # Spin register as a list without u,v spins
    iter_qubits = [j for j in range(0, min(u, v))] + [j for j in range(
        min(u, v)+1, max(u, v))] + [j for j in range(max(u, v)+1, n_qubits)]

    # Initialize products
    prod1 = s**2/2 * np.cos(2*gamma*(h_u - h_v))
    prod2 = -s**2/2 * np.cos(2*gamma*(h_u + h_v))
    prod3 = -c*s*np.sin(2*gamma*J_uv) * np.cos(2*gamma*h_u)
    prod4 = -c*s*np.sin(2*gamma*J_uv) * np.cos(2*gamma*h_v)

    # Loop over edges connecting u and v to other spins
    for n in iter_qubits:

        # Edges between u,v and another spin in the register
        edge1 = tuple([min(u, n), max(u, n)])
        edge2 = tuple([min(v, n), max(v, n)])

        # If edge not present in the graph the associated weight is set to 0
        J_un = 0 if hamil_graph.get(edge1) is None else hamil_graph[edge1]
        J_vn = 0 if hamil_graph.get(edge2) is None else hamil_graph[edge2]

        # Add factor to the products
        prod1 *= np.cos(2*gamma*(J_un - J_vn))
        prod2 *= np.cos(2*gamma*(J_un + J_vn))
        prod3 *= np.cos(2*gamma*J_un)
        prod4 *= np.cos(2*gamma*J_vn)

    # Add the contribution from each product term
    corr = prod1 + prod2 + prod3 + prod4

    return corr


def energy_expectation_analytical(angles:Union[list,tuple],hamiltonian:Hamiltonian):
    """
    Computes the expectation value of the Hamiltonian for an analytical expression.

    NOTE: Only valid for single layer QAOA Ansatz with X mixer Hamiltonian and classical
    Hamiltonians with up to quadratic terms.

    Parameters
    ----------
    angles: `list` or `tuple`
        QAOA angles at which the Hamiltonian expectation value is computed
    hamiltonian: `Hamiltonian`
        Classical Hamiltonian from which the expectation value is computed.
    """
    
    # Extract terms and coefficients from the Hamiltonian
    terms = [pauli_term.qubit_indices for pauli_term in hamiltonian.terms]
    coeffs = hamiltonian.coeffs
    
    energy = 0
    
    # Compute the expectation value of each term and add its local energy contribution
    for coeff,term in zip(coeffs,terms):
        
        if len(term) == 2:
        
            local_energy = exp_val_pair_analytical(term,hamiltonian,angles)
            
        else:
            
            local_energy = exp_val_single_analytical(term[0],hamiltonian,angles)
        
        energy += coeff * local_energy
    
    # Add constant shift contribution
    energy += hamiltonian.constant
    
    return energy


def ring_of_disagrees(reg: List[int]) -> Hamiltonian:
    """
    Builds the cost Hamiltonian for the "Ring of Disagrees".

    Parameters
    ----------
    reg: `list`
        register of qubits in the system.

    Returns
    -------
    ring_hamil: `Hamiltonian`
        Hamiltonian object containing Ring of Disagrees model.

    Notes
    -----
    This model is introduced in https://arxiv.org/abs/1411.4028
    """

    # Number of qubits from input register
    n_qubits = len(reg)

    # Define terms for the ring structure
    terms = [(reg[i], reg[(i+1) % n_qubits]) for i in range(n_qubits)]

    # Define coefficients as in original formulation of the model
    coeffs = [0.5]*len(terms)

    # Constant term as in original formulation of the model
    constant = -len(terms)*0.5

    # Define Hamiltonian
    ring_hamil = Hamiltonian.classical_hamiltonian(terms, coeffs,
                                                   constant=constant)
    return ring_hamil


################################################################################
# OTHER MISCELLANEOUS
################################################################################


def flip_counts(counts_dictionary: dict) -> dict:
    """
    Returns a counts/probability dictionary that have their keys flipped. This
    formats the bit-strings from a right-most bit representing being the first 
    qubit to the left-most bit representing the first qubit.

    Parameters
    ----------
    counts_dictionary: `dict`
        Count dictionary whose keys are flipped.
        
    Returns
    -------
    output_counts_dictionary: `dict`
        Count dictionary with flipped keys.
    """

    output_counts_dictionary = dict()

    for key, value in counts_dictionary.items():
        output_counts_dictionary[key[::-1]] = value

    return output_counts_dictionary

def qaoa_probabilities(statevector) -> dict:
    """
    Return a qiskit-style probability dictionary from a statevector.

    Parameters
    ----------
    statevector: `np.ndarray[complex]`
        The wavefunction whose probability distribution needs to be calculated.

    Returns
    -------
    prob_dict: `dict`
        Probabilities represented as a python dictionary with basis states stored
        as keys and their probabilities as their corresponding values.
    """
    # Define list of probabilities from wavefunction amplitudes
    prob_vec = np.real(np.conjugate(statevector)*statevector)

    # Extract number of qubits from size of probability
    n_qubits = int(np.log2(len(prob_vec)))

    # Initialize probability dictionary
    prob_dict = {}

    for x in range(len(prob_vec)):

        # Define binary representation of each state, with qubit-0 most significant bit
        key = np.binary_repr(x, n_qubits)[::-1]

        # Update probability dictionary
        prob_dict.update({key: prob_vec[x]})

    return prob_dict



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import Counter
import numpy as np
from pyquil import Program, gates, quilbase

from ...basebackend import QAOABaseBackendShotBased, QAOABaseBackendCloud, QAOABaseBackendParametric
from ...qaoa_parameters.baseparams import QAOACircuitParams, QAOAVariationalBaseParams
from ...devices import DevicePyquil
from ...qaoa_parameters.gatemap import RZZGateMap, SWAPGateMap

def check_edge_connectivity(executable: Program, device: DevicePyquil):

    '''
    Check that the program does not contain 2-qubit terms that is not present in the QPU's topology (to prevent quilc from crashing).
    
    Parameters
    ----------
    executable: `Program`
        pyQuil executable program.
    device: `DevicePyquil`
        An object of the class ``DevicePyquil`` which contains information on pyQuil's `QuantumComputer` object, used to extract the selected QPU's topology.
    
    Returns
    -------
        None
    """
    '''

    qpu_graph = device.quantum_computer.qubit_topology()
    
    instrs = [instr for instr in executable if type(instr) == quilbase.Gate]
    pair_instrs = [list(instr.get_qubits()) for instr in instrs if len(instr.get_qubits()) == 2]

    for term in pair_instrs:
        if len(term) == 2:

            assert term in qpu_graph.edges(), f"Term {term} is not an edge on the QPU graph of {device.device_name}."
    
    

class QAOAPyQuilQPUBackend(QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased):
    """
    A QAOA backend object for real Rigetti QPUs

    Parameters
    ----------
    device: `DevicePyquil`
        The device object to access pyquil devices with credentials.
    circuit_params: `QAOACircuitParams`
        An object of the class ``QAOACircuitParams`` which contains information on 
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `pyquil.Program`
        The state prepended to the circuit.
    append_state: `pyquil.Program`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the QAOA part of the circuit.
    cvar_alpha: `float`
        Conditional Value-at-Risk (CVaR) – a measure that takes into account only the tail of the
        probability distribution arising from the circut's count dictionary. Must be between 0 and 1. Check
        https://arxiv.org/abs/1907.04769 for further details.
    active_reset: 
        Whether to use the pyQuil's active reset scheme to reset qubits between shots. 
    rewiring:
        Rewiring scheme to be used for Pyquil. 
        Either PRAGMA INITIAL_REWIRING "NAIVE" or PRAGMA INITIAL_REWIRING "PARTIAL". 
        If None, pyquil defaults according to:
        NAIVE: The qubits used in all instructions in the program satisfy the topological constraints of the device.
        PARTIAL: Otherwise.
    """

    def __init__(self,
                 device: DevicePyquil,
                 circuit_params: QAOACircuitParams,
                 n_shots: int,
                 prepend_state: Program,
                 append_state: Program,
                 init_hadamard: bool,
                 cvar_alpha: float,
                 active_reset: bool = False,
                 rewiring: str = '',
                 qubit_layout: list = []
                 ):

        QAOABaseBackendShotBased.__init__(self,
                                          circuit_params,
                                          n_shots,
                                          prepend_state,
                                          append_state,
                                          init_hadamard,
                                          cvar_alpha)
        QAOABaseBackendCloud.__init__(self, device)

        self.active_reset = active_reset
        self.rewiring = rewiring
        self.qureg = self.circuit_params.qureg

        # self.qureg_placeholders = QubitPlaceholder.register(self.n_qubits)
        self.qubit_layout = self.qureg if qubit_layout == [] else qubit_layout
        self.qubit_mapping = dict(zip(self.qureg, self.qubit_layout))
        
        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.get_qubits()), "Cannot attach a bigger circuit " \
                                                                "to the QAOA routine"

        self.parametric_circuit = self.parametric_qaoa_circuit
        native_prog = self.device.quantum_computer.compiler.quil_to_native_quil(
            self.parametric_circuit)
        self.prog_exe = self.device.quantum_computer.compiler.native_quil_to_executable(
            native_prog)
        
        # Check program connectivity against QPU connectivity
        # TODO: reconcile with PRAGMA PRESERVE

        # check_edge_connectivity(self.prog_exe, device)


    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Program:
        """
        Injects angles into created executable parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
            A pyquil.Program (executable) object.
        """
        angles_list = np.array(self.obtain_angles_for_pauli_list(
            self.abstract_circuit, params), dtype=float)
        angle_declarations = list(self.prog_exe.declarations.keys())
        angle_declarations.remove('ro')
        for i, param_name in enumerate(angle_declarations):
            self.prog_exe.write_memory(
                region_name=param_name, value=angles_list[i])

        return self.prog_exe

    @property
    def parametric_qaoa_circuit(self) -> Program:
        """
        Creates a parametric QAOA circuit (pyquil.Program object), given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit. 

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
        
        Returns
        -------
        `pyquil.Program`
            A pyquil.Program object.
        """
        if self.active_reset:
            parametric_circuit = Program(gates.RESET())
        else:
            parametric_circuit = Program()
        
        if self.rewiring != None:
            if self.rewiring in ['PRAGMA INITIAL_REWIRING "NAIVE"', 'PRAGMA INITIAL_REWIRING "PARTIAL"', '']:
                parametric_circuit += Program(self.rewiring)
            else:
                raise ValueError('Rewiring command not recognized. Please use ''PRAGMA INITIAL_REWIRING "NAIVE"'' or ''PRAGMA INITIAL_REWIRING "PARTIAL"''')
        
        # declare the read-out register
        ro = parametric_circuit.declare('ro', 'BIT', self.n_qubits)

        if self.prepend_state:
            parametric_circuit += self.prepend_state
            
        # Initial state is all |+>
        if self.init_hadamard:
            for i in self.qureg:
                parametric_circuit += gates.RZ(np.pi, self.qubit_mapping[i]) 
                parametric_circuit += gates.RX(np.pi/2, self.qubit_mapping[i]) 
                parametric_circuit += gates.RZ(np.pi/2, self.qubit_mapping[i]) 
                parametric_circuit += gates.RX(-np.pi/2, self.qubit_mapping[i]) 

        # create a list of gates in order of application on quantum circuit
        for each_gate in self.abstract_circuit:
            gate_label = ''.join(str(label) for label in each_gate.pauli_label)
            angle_param = parametric_circuit.declare(
                f'pauli{gate_label}', 'REAL', 1)
            each_gate.rotation_angle = angle_param
            if isinstance(each_gate, RZZGateMap) or isinstance(each_gate, SWAPGateMap):
                decomposition = each_gate.decomposition('standard2')
            else:
                decomposition = each_gate.decomposition('standard')
                
            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0]()
                qubits, rotation_angle = each_tuple[1]
                if isinstance(qubits,list):
                    new_qubits = [self.qubit_mapping[qubit] for qubit in qubits]
                else:
                    new_qubits = self.qubit_mapping[qubits]
                parametric_circuit = gate.apply_pyquil_gate(new_qubits,rotation_angle,parametric_circuit)

        if self.append_state:
            parametric_circuit += self.append_state
            
        # Measurement instructions
        for i, qubit in enumerate(self.qureg):
            parametric_circuit += gates.MEASURE(self.qubit_mapping[qubit], ro[i])

        parametric_circuit.wrap_in_numshots_loop(self.n_shots)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams) -> dict:
        """
        Execute the circuit and obtain the counts.

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing 
            variable parameters.

        Returns
        -------
        counts : dictionary
            A dictionary with the bitstring as the key and the number of counts as its value.
        """
        executable_program = self.qaoa_circuit(params)

        result = self.device.quantum_computer.run(executable_program)

        # TODO: check the endian (big or little) ordering of measurement outcomes
        meas_list = [''.join(str(bit) for bit in bitstring)
                     for bitstring in result.readout_data['ro']]
        
        # Expose counts
        counts = Counter(list(meas_list))
        self.measurement_outcomes = counts
        return counts

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the pyQuil program to a OpenQASM string.
        """
        raise NotImplementedError()
        # qasm_program = self.device.quantum_computer.compiler.quil_to_qasm(self.qaoa_circuit(params))
        # return qasm_program

    def reset_circuit(self):
        """
        Reset self.program after performing a computation. Also handle active reset and rewirings.
        """
        pass



#   Copyright 2022 Entropica Labs
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Different hyper-parameterisations and trainable-parameterisations for QAOA

Hyper-Parameters help with problem hamiltonian initialisation and fixing `p`
for the QAOA problem

AbstractParams and child classes let one choose the parameterisation for
solving the QAOA problem in hand.  
"""

from .standardparams import *
from .extendedparams import *
from .fourierparams import *
from .annealingparams import *
from .qaoa_params import *
from .operators import *
from .baseparams import *
from .gatemap import *
from .gates import *
from .variational_params_converters import converter